"""sonusai torchl_onnx

usage: torchl_onnx [-hv] [-b BATCH] [-t TSTEPS] [-o OUTPUT] MODEL CKPT

options:
    -h, --help
    -v, --verbose                   Be verbose
    -b BATCH, --batch BATCH         Batch size. [default: 1]
    -t TSTEPS, --tsteps TSTEPS      Timesteps
    -o OUTPUT, --output OUTPUT      Output directory

Convert a trained PL (Pytorch Lightning) model to ONNX. The model is specified as an
sctl_*.py model file (sctl: sonusai custom torch lightning) and a checkpoint file
for loading weights.

Inputs:
    MODEL       SonusAI Python custom model file.
    CKPT        A PL checkpoint file
    BATCH       Specify a batch size, else defaults to 1.
    TSTEPS      Specify a timestep dimension size, if the model has a timestep dimension.

Outputs:
    OUTPUT/     A directory containing:
                    <CKPT>.onnx        Model file with batch_size and timesteps equal to provided parameters
                    <CKPT>-b1.onnx     Model file with batch_size=1 and if the timesteps dimension exists it
                                       is set to 1 (useful for real-time inference applications)
                    torchl_onnx.log

Results are written into subdirectory <MODEL>-<TIMESTAMP> unless OUTPUT is specified.

"""
from sonusai import logger


def main() -> None:
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    from os import makedirs
    from os.path import basename
    from os.path import splitext

    from sonusai import logger
    from sonusai.utils import import_keras_model

    verbose = args['--verbose']
    batch_size = args['--batch']
    timesteps = args['--tsteps']
    model_path = args['MODEL']
    ckpt_path = args['CKPT']
    output_dir = args['--output']

    # Import model definition file first to check
    model_base = basename(model_path)
    model_root = splitext(model_base)[0]
    logger.info(f'Importing model from {model_base}')
    try:
        litemodule = import_keras_model(model_path)  # note works for PL as well as Keras
    except Exception as e:
        logger.exception(f'Error: could not import model from {model_path}: {e}')
        raise SystemExit(1)

    # Load checkpoint first to get hparams if available
    from torch import load as load
    ckpt_base = basename(ckpt_path)
    ckpt_root = splitext(ckpt_base)[0]
    logger.info(f'Loading checkpoint from {ckpt_base}')
    try:
        checkpoint = load(ckpt_path, map_location=lambda storage, loc: storage)
    except Exception as e:
        logger.exception(f'Error: could not load checkpoint from {ckpt_path}: {e}')
        raise SystemExit(1)

    from os.path import join, isdir, dirname, exists
    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler

    from torchinfo import summary

    if batch_size is not None:
        batch_size = int(batch_size)
    if batch_size != 1:
        batch_size = 1
        logger.info(f'For now prediction only supports batch_size = 1, forcing it to 1 now')

    if timesteps is not None:
        timesteps = int(timesteps)

    if output_dir is None:
        output_dir = dirname(ckpt_path)
    else:
        if not isdir(output_dir):
            makedirs(output_dir, exist_ok=True)

    ofname = join(output_dir, ckpt_root + '.onnx')
    # First try, then add date
    if exists(ofname):
        # add hour-min-sec if necessary
        from datetime import datetime
        ts = datetime.now()
        ofname = join(output_dir, ckpt_root + '-' + ts.strftime('%Y%m%d') + '.onnx')
    ofname_root = splitext(ofname)[0]

    # Setup logging file
    create_file_handler(ofname_root + '-onnx.log')
    update_console_handler(verbose)
    initial_log_messages('torchl_onnx')
    logger.info(f'Imported model from {model_base}')
    logger.info(f'Loaded checkpoint from {ckpt_base}')

    if 'hyper_parameters' in checkpoint:
        hparams = checkpoint['hyper_parameters']
        logger.info(f'Found hyper-params on checkpoint named {checkpoint["hparams_name"]} '
                    f'with {len(hparams)} total hparams.')
        if batch_size is not None and hparams['batch_size'] != batch_size:
            if batch_size != 1:
                batch_size = 1
                logger.info(f'For now prediction only supports batch_size = 1, forcing it to 1 now')
            logger.info(f'Overriding batch_size: default = {hparams["batch_size"]}; specified = {batch_size}.')
            hparams["batch_size"] = batch_size

        if timesteps is not None:
            if hparams['timesteps'] == 0 and timesteps != 0:
                logger.warning(f'Model does not contain timesteps; ignoring override.')
                timesteps = 0

            if hparams['timesteps'] != 0 and timesteps == 0:
                logger.warning(f'Model contains timesteps; ignoring override, using model default.')
                timesteps = hparams['timesteps']

            if hparams['timesteps'] != timesteps:
                logger.info(f'Overriding timesteps: default = {hparams["timesteps"]}; specified = {timesteps}.')
                hparams['timesteps'] = timesteps

        logger.info(f'Building model with hparams and batch_size={batch_size}, timesteps={timesteps}')
        try:
            model = litemodule.MyHyperModel(**hparams)  # use hparams
            # litemodule.MyHyperModel.load_from_checkpoint(ckpt_name, **hparams)
        except Exception as e:
            logger.exception(f'Error: model build (MyHyperModel) in {model_base} failed: {e}')
            raise SystemExit(1)
    else:
        logger.info(f'Warning: found checkpoint with no hyper-parameters, building model with defaults')
        try:
            tmp = litemodule.MyHyperModel()  # use default hparams
        except Exception as e:
            logger.exception(f'Error: model build (MyHyperModel) in {model_base} failed: {e}')
            raise SystemExit(1)

        if batch_size is not None:
            if tmp.batch_size != batch_size:
                logger.info(f'Overriding batch_size: default = {tmp.batch_size}; specified = {batch_size}.')
        else:
            batch_size = tmp.batch_size  # inherit

        if timesteps is not None:
            if tmp.timesteps == 0 and timesteps != 0:
                logger.warning(f'Model does not contain timesteps; ignoring override.')
                timesteps = 0

            if tmp.timesteps != 0 and timesteps == 0:
                logger.warning(f'Model contains timesteps; ignoring override.')
                timesteps = tmp.timesteps

            if tmp.timesteps != timesteps:
                logger.info(f'Overriding timesteps: default = {tmp.timesteps}; specified = {timesteps}.')
        else:
            timesteps = tmp.timesteps

        logger.info(f'Building model with default hparams and batch_size= {batch_size}, timesteps={timesteps}')
        model = litemodule.MyHyperModel(timesteps=timesteps, batch_size=batch_size)

    logger.info('')
    logger.info(summary(model))
    logger.info('')
    logger.info(f'feature       {model.hparams.feature}')
    logger.info(f'num_classes   {model.num_classes}')
    logger.info(f'batch_size    {model.hparams.batch_size}')
    logger.info(f'timesteps     {model.hparams.timesteps}')
    logger.info(f'flatten       {model.flatten}')
    logger.info(f'add1ch        {model.add1ch}')
    logger.info(f'truth_mutex   {model.truth_mutex}')
    logger.info(f'input_shape   {model.input_shape}')
    logger.info('')
    logger.info(f'Loading weights from {ckpt_base}')
    # model = model.load_from_checkpoint(ckpt_path)     # weights only, has problems - needs investigation
    model.load_state_dict(checkpoint["state_dict"])
    model.eval()
    logger.info(f'Creating onnx model ...')
    model.to_onnx(file_path=ofname)  # , input_sample=model.input_shape)


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        exit()



from sonusai.mixture.datatypes import AugmentationRules
from sonusai.mixture.datatypes import AugmentedTarget
from sonusai.mixture.datatypes import AugmentedTargets
from sonusai.mixture.datatypes import TargetFile
from sonusai.mixture.datatypes import TargetFiles


def get_augmented_targets(target_files: TargetFiles,
                          target_augmentations: AugmentationRules,
                          mixups: list[int] = None) -> AugmentedTargets:
    from .augmentation import get_augmentation_indices_for_mixup
    from .augmentation import get_mixups

    if mixups is None:
        mixups = get_mixups(target_augmentations)

    augmented_targets: AugmentedTargets = []
    for mixup in mixups:
        augmentation_indices = get_augmentation_indices_for_mixup(target_augmentations, mixup)
        for target_index in range(len(target_files)):
            for augmentation_index in augmentation_indices:
                augmented_targets.append(AugmentedTarget(target_id=target_index,
                                                         target_augmentation_id=augmentation_index))

    return augmented_targets


def get_truth_indices_for_target(target: TargetFile) -> list[int]:
    """Get a list of truth indices for a given target."""
    index = [truth_setting.index for truth_setting in target.truth_settings]

    # flatten, uniquify, and sort
    return sorted(list(set([item for sublist in index for item in sublist])))


def get_truth_indices_for_augmented_target(augmented_target: AugmentedTarget, targets: TargetFiles) -> list[int]:
    return get_truth_indices_for_target(targets[augmented_target.target_id])


def get_mixup_for_augmented_target(augmented_target: AugmentedTarget, augmentations: AugmentationRules) -> int:
    return augmentations[augmented_target.target_augmentation_id].mixup


def get_target_ids_for_truth_index(targets: TargetFiles,
                                   truth_index: int,
                                   allow_multiple: bool = False) -> list[int]:
    """Get a list of target indices containing the given truth index.

    If allow_multiple is True, then include targets that contain multiple truth indices.
    """
    target_indices = set()
    for target_index, target in enumerate(targets):
        indices = get_truth_indices_for_target(target)
        if len(indices) == 1 or allow_multiple:
            for index in indices:
                if index == truth_index + 1:
                    target_indices.add(target_index)

    return sorted(list(target_indices))


def get_augmented_target_ids_for_truth_index(augmented_targets: AugmentedTargets,
                                             targets: TargetFiles,
                                             augmentations: AugmentationRules,
                                             truth_index: int,
                                             mixup: int,
                                             allow_multiple: bool = False) -> list[int]:
    """Get a list of augmented target indices containing the given truth index.

    If allow_multiple is True, then include targets that contain multiple truth indices.
    """
    augmented_target_ids = set()
    for augmented_target_id, augmented_target in enumerate(augmented_targets):
        if get_mixup_for_augmented_target(augmented_target=augmented_target, augmentations=augmentations) == mixup:
            indices = get_truth_indices_for_augmented_target(augmented_target=augmented_target, targets=targets)
            if len(indices) == 1 or allow_multiple:
                for index in indices:
                    if index == truth_index + 1:
                        augmented_target_ids.add(augmented_target_id)

    return sorted(list(augmented_target_ids))


def get_augmented_target_ids_by_class(augmented_targets: AugmentedTargets,
                                      targets: TargetFiles,
                                      target_augmentations: AugmentationRules,
                                      mixup: int,
                                      num_classes: int,
                                      truth_mutex: bool) -> list[list[int]]:
    if truth_mutex:
        num_classes -= 1

    indices = []
    for idx in range(num_classes):
        indices.append(
            get_augmented_target_ids_for_truth_index(augmented_targets=augmented_targets,
                                                     targets=targets,
                                                     augmentations=target_augmentations,
                                                     truth_index=idx,
                                                     mixup=mixup))
    return indices


def get_target_augmentations_for_mixup(target_augmentations: AugmentationRules, mixup: int) -> AugmentationRules:
    """Get target augmentations for a given mixup value

    :param target_augmentations: List of target augmentation rules
    :param mixup: Mixup value of interest
    :return: Target augmentations
    """
    return [target_augmentation for target_augmentation in target_augmentations if target_augmentation.mixup == mixup]


def get_augmented_target_ids_for_mixup(augmented_targets: AugmentedTargets,
                                       targets: TargetFiles,
                                       target_augmentations: AugmentationRules,
                                       mixup: int,
                                       num_classes: int,
                                       truth_mutex: bool) -> list[list[int]]:
    from collections import deque
    from random import shuffle

    from sonusai import SonusAIError

    mixup_indices = []

    if mixup == 1:
        for index, augmented_target in enumerate(augmented_targets):
            if get_mixup_for_augmented_target(augmented_target=augmented_target,
                                              augmentations=target_augmentations) == 1:
                mixup_indices.append([index])
        return mixup_indices

    augmented_target_ids_by_class = get_augmented_target_ids_by_class(augmented_targets=augmented_targets,
                                                                      targets=targets,
                                                                      target_augmentations=target_augmentations,
                                                                      mixup=mixup,
                                                                      num_classes=num_classes,
                                                                      truth_mutex=truth_mutex)

    if mixup > num_classes:
        raise SonusAIError(
            f'Specified mixup, {mixup}, is greater than the number of classes, {num_classes}')

    de: deque[int] = deque()

    # Keep looping until not enough targets remain for mixup
    while sum([1 for x in augmented_target_ids_by_class if x]) >= mixup:
        # Need more class indices?
        if len(de) < mixup:
            # Only choose classes that still have data
            counts = [len(item) for item in augmented_target_ids_by_class]
            # Need to subtract out indices already in the deque
            for idx in de:
                counts[idx] -= 1
            indices = [idx for idx, val in enumerate(counts) if val > 0]
            shuffle(indices)
            # Keep shuffling if the deque is not empty and the first new index matches the last item
            # (so that a class does not appear twice in a mixup)
            while de and indices[0] == de[-1]:
                shuffle(indices)
            for index in indices:
                de.append(index)

        class_indices = [de.popleft() for _ in range(mixup)]

        target_indices = []
        for class_index in class_indices:
            target_indices.append(augmented_target_ids_by_class[class_index].pop())

        mixup_indices.append(target_indices)

    return mixup_indices



# SonusAI truth functions
from .crm import crm
from .crm import crmp
from .energy import energy_f
from .energy import energy_t
from .energy import mapped_snr_f
from .energy import snr_f
from .file import file
from .phoneme import phoneme
from .sed import sed
from .target import target_f
from .target import target_mixture_f
from .target import target_swin_f



from typing import Any
from typing import Callable

from sonusai.mixture.datatypes import GeneralizedIDs
from sonusai.mixture.mixdb import MixtureDatabase


def get_mixids_from_mixture_field_predicate(mixdb: MixtureDatabase,
                                            field: str,
                                            mixids: GeneralizedIDs = None,
                                            predicate: Callable[[Any], bool] = None) -> dict[int, list[int]]:
    """
    Generate mixture IDs based on mixture field and predicate
    Return a dictionary where:
        - keys are the matching field values
        - values are lists of the mixids that match the criteria
    """
    mixid_out = mixdb.mixids_to_list(mixids)

    if predicate is None:
        def predicate(_: Any) -> bool:
            return True

    criteria_set = set()
    for m_id in mixid_out:
        value = getattr(mixdb.mixture(m_id), field)
        if isinstance(value, list):
            for v in value:
                if predicate(v):
                    criteria_set.add(v)
        elif predicate(value):
            criteria_set.add(value)
    criteria = sorted(list(criteria_set))

    result: dict[int, list[int]] = {}
    for criterion in criteria:
        result[criterion] = []
        for m_id in mixid_out:
            value = getattr(mixdb.mixture(m_id), field)
            if isinstance(value, list):
                for v in value:
                    if v == criterion:
                        result[criterion].append(m_id)
            elif value == criterion:
                result[criterion].append(m_id)

    return result


def get_mixids_from_truth_settings_field_predicate(mixdb: MixtureDatabase,
                                                   field: str,
                                                   mixids: GeneralizedIDs = None,
                                                   predicate: Callable[[Any], bool] = None) -> dict[int, list[int]]:
    """
    Generate mixture IDs based on target truth_settings field and predicate
    Return a dictionary where:
        - keys are the matching field values
        - values are lists of the mixids that match the criteria
    """
    mixid_out = mixdb.mixids_to_list(mixids)

    # Get all field values
    values = get_all_truth_settings_values_from_field(mixdb, field)

    if predicate is None:
        def predicate(_: Any) -> bool:
            return True

    # Get only values of interest
    values = [value for value in values if predicate(value)]

    result = {}
    for value in values:
        # Get a list of targets for each field value
        indices = []
        for t_id in mixdb.target_file_ids:
            target = mixdb.target_file(t_id)
            for truth_setting in target.truth_settings:
                if value in getattr(truth_setting, field):
                    indices.append(t_id)
        indices = sorted(list(set(indices)))

        mixids = []
        for index in indices:
            for m_id in mixid_out:
                if index in [target.file_id for target in mixdb.mixture(m_id).targets]:
                    mixids.append(m_id)

        mixids = sorted(list(set(mixids)))
        if mixids:
            result[value] = mixids

    return result


def get_all_truth_settings_values_from_field(mixdb: MixtureDatabase, field: str) -> list:
    """
    Generate a list of all values corresponding to the given field in truth_settings
    """
    result = []
    for target in mixdb.target_files:
        for truth_setting in target.truth_settings:
            value = getattr(truth_setting, field)
            if isinstance(value, str):
                value = [value]
            result.extend(value)

    return sorted(list(set(result)))


def get_mixids_from_noise(mixdb: MixtureDatabase,
                          mixids: GeneralizedIDs = None,
                          predicate: Callable[[Any], bool] = None) -> dict[int, list[int]]:
    """
    Generate mixids based on noise index predicate
    Return a dictionary where:
        - keys are the noise indices
        - values are lists of the mixids that match the noise index
    """
    return get_mixids_from_mixture_field_predicate(mixdb=mixdb,
                                                   mixids=mixids,
                                                   field='noise_id',
                                                   predicate=predicate)


def get_mixids_from_target(mixdb: MixtureDatabase,
                           mixids: GeneralizedIDs = None,
                           predicate: Callable[[Any], bool] = None) -> dict[int, list[int]]:
    """
    Generate mixids based on a target index predicate
    Return a dictionary where:
        - keys are the target indices
        - values are lists of the mixids that match the target index
    """
    return get_mixids_from_mixture_field_predicate(mixdb=mixdb,
                                                   mixids=mixids,
                                                   field='target_id',
                                                   predicate=predicate)


def get_mixids_from_snr(mixdb: MixtureDatabase,
                        mixids: GeneralizedIDs = None,
                        predicate: Callable[[Any], bool] = None) -> dict[float, list[int]]:
    """
    Generate mixids based on an SNR predicate
    Return a dictionary where:
        - keys are the SNRs
        - values are lists of the mixids that match the SNR
    """
    from typing import Any

    mixid_out = mixdb.mixids_to_list(mixids)

    # Get all the SNRs
    snrs = [float(snr) for snr in mixdb.all_snrs if not snr.is_random]

    if predicate is None:
        def predicate(_: Any) -> bool:
            return True

    # Get only the SNRs of interest (filter on predicate)
    snrs = [snr for snr in snrs if predicate(snr)]

    result = {}
    for snr in snrs:
        # Get a list of mixids for each SNR
        result[snr] = sorted(
            [i for i, mixture in enumerate(mixdb.mixtures) if mixture.snr == snr and i in mixid_out])

    return result


def get_mixids_from_truth_index(mixdb: MixtureDatabase,
                                mixids: GeneralizedIDs = None,
                                predicate: Callable[[Any], bool] = None) -> dict[int, list[int]]:
    """
    Generate mixids based on a truth index predicate
    Return a dictionary where:
        - keys are the truth indices
        - values are lists of the mixids that match the truth index
    """
    return get_mixids_from_truth_settings_field_predicate(mixdb=mixdb,
                                                          mixids=mixids,
                                                          field='index',
                                                          predicate=predicate)


def get_mixids_from_truth_function(mixdb: MixtureDatabase,
                                   mixids: GeneralizedIDs = None,
                                   predicate: Callable[[Any], bool] = None) -> dict[int, list[int]]:
    """
    Generate mixids based on a truth function predicate
    Return a dictionary where:
        - keys are the truth functions
        - values are lists of the mixids that match the truth function
    """
    return get_mixids_from_truth_settings_field_predicate(mixdb=mixdb,
                                                          mixids=mixids,
                                                          field='function',
                                                          predicate=predicate)



import numpy as np


def calc_pesq(hypothesis: np.ndarray, reference: np.ndarray, error_value: float = 0.0) -> float:
    """Computes the PESQ score of speech estimate audio vs. the clean speech estimate audio

    Upon error, assigns a value of 0, or user specified value in error_value

    :param hypothesis: speech estimated audio
    :param reference: speech reference audio
    :param error_value:
    :return: value between -0.5 to 4.5
    """
    import warnings

    from pesq import pesq

    from sonusai import logger
    from sonusai.mixture import SAMPLE_RATE

    try:
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            score = pesq(SAMPLE_RATE, reference, hypothesis, mode='wb')
    except Exception as e:
        logger.debug(f'PESQ error {e}')
        score = error_value

    return score



from dataclasses import dataclass
from dataclasses import field
from typing import Optional
from typing import TypeAlias

import numpy as np
import numpy.typing as npt
from dataclasses_json import DataClassJsonMixin

AudioT: TypeAlias = npt.NDArray[np.float32]
AudiosT: TypeAlias = list[AudioT]

ListAudiosT: TypeAlias = list[AudiosT]

Truth: TypeAlias = npt.NDArray[np.float32]
Segsnr: TypeAlias = npt.NDArray[np.float32]

AudioF: TypeAlias = npt.NDArray[np.complex64]
AudiosF: TypeAlias = list[AudioF]

EnergyT: TypeAlias = npt.NDArray[np.float32]
EnergyF: TypeAlias = npt.NDArray[np.float32]

Feature: TypeAlias = npt.NDArray[np.float32]

Predict: TypeAlias = npt.NDArray[np.float32]

# Json type defined to maintain compatibility with DataClassJsonMixin
Json: TypeAlias = dict | list | str | int | float | bool | None


class DataClassSonusAIMixin(DataClassJsonMixin):
    def __str__(self):
        return f'{self.to_dict()}'

    # Override DataClassJsonMixin to remove dictionary keys with values of None
    def to_dict(self, encode_json=False) -> dict[str, Json]:
        def del_none(d):
            if isinstance(d, dict):
                for key, value in list(d.items()):
                    if value is None:
                        del d[key]
                    elif isinstance(value, dict):
                        del_none(value)
                    elif isinstance(value, list):
                        for item in value:
                            del_none(item)
            elif isinstance(d, list):
                for item in d:
                    del_none(item)
            return d

        return del_none(super().to_dict(encode_json))


@dataclass(frozen=True)
class TruthSetting(DataClassSonusAIMixin):
    config: Optional[dict] = None
    function: Optional[str] = None
    index: Optional[list[int]] = None

    def __hash__(self):
        return hash(self.to_json())

    def __eq__(self, other):
        return isinstance(other, TruthSetting) and hash(self) == hash(other)


TruthSettings: TypeAlias = list[TruthSetting]
NumberStr: TypeAlias = float | int | str
OptionalNumberStr: TypeAlias = Optional[NumberStr]
OptionalListNumberStr: TypeAlias = Optional[list[NumberStr]]
EQ: TypeAlias = tuple[float | int, float | int, float | int]


@dataclass
class AugmentationRule(DataClassSonusAIMixin):
    normalize: OptionalNumberStr = None
    pitch: OptionalNumberStr = None
    tempo: OptionalNumberStr = None
    gain: OptionalNumberStr = None
    eq1: OptionalListNumberStr = None
    eq2: OptionalListNumberStr = None
    eq3: OptionalListNumberStr = None
    lpf: OptionalNumberStr = None
    ir: OptionalNumberStr = None
    mixup: Optional[int] = 1


AugmentationRules: TypeAlias = list[AugmentationRule]


@dataclass
class Augmentation(DataClassSonusAIMixin):
    normalize: Optional[float] = None
    pitch: Optional[float] = None
    tempo: Optional[float] = None
    gain: Optional[float] = None
    eq1: Optional[EQ] = None
    eq2: Optional[EQ] = None
    eq3: Optional[EQ] = None
    lpf: Optional[float] = None
    ir: Optional[int] = None


Augmentations: TypeAlias = list[Augmentation]


@dataclass(frozen=True)
class UniversalSNRGenerator:
    is_random: bool
    _raw_value: float | str

    @property
    def value(self) -> float:
        if self.is_random:
            from .augmentation import evaluate_random_rule

            return float(evaluate_random_rule(str(self._raw_value)))

        return float(self._raw_value)


class UniversalSNR(float):
    def __new__(cls, value: float, is_random: bool = False):
        return float.__new__(cls, value)

    def __init__(self, value: float, is_random: bool = False) -> None:
        float.__init__(value)
        self._is_random = bool(is_random)

    @property
    def is_random(self) -> bool:
        return self._is_random


# @dataclass(frozen=True)
# class UniversalSNR:
#     is_random: bool
#     value: float
#
#     def __lt__(self, other) -> bool:
#         return self.value < other.value


@dataclass
class TargetFile(DataClassSonusAIMixin):
    name: str
    samples: int
    truth_settings: TruthSettings
    class_balancing_augmentation: Optional[AugmentationRule] = None
    level_type: Optional[str] = None

    @property
    def duration(self) -> float:
        from .constants import SAMPLE_RATE

        return self.samples / SAMPLE_RATE


TargetFiles: TypeAlias = list[TargetFile]


@dataclass
class AugmentedTarget(DataClassSonusAIMixin):
    target_id: int
    target_augmentation_id: int


AugmentedTargets: TypeAlias = list[AugmentedTarget]


@dataclass
class NoiseFile(DataClassSonusAIMixin):
    name: str
    samples: int

    @property
    def duration(self) -> float:
        from .constants import SAMPLE_RATE

        return self.samples / SAMPLE_RATE


NoiseFiles: TypeAlias = list[NoiseFile]
ClassCount: TypeAlias = list[int]

GeneralizedIDs: TypeAlias = str | int | list[int] | range


@dataclass(frozen=True)
class TruthFunctionConfig(DataClassSonusAIMixin):
    feature: str
    mutex: bool
    num_classes: int
    target_gain: float
    config: Optional[dict] = None
    function: Optional[str] = None
    index: Optional[list[int]] = None


@dataclass
class GenMixData:
    targets: Optional[AudiosT] = None
    target: Optional[AudioT] = None
    noise: Optional[AudioT] = None
    mixture: Optional[AudioT] = None
    truth_t: Optional[Truth] = None
    segsnr_t: Optional[Segsnr] = None


@dataclass
class GenFTData:
    feature: Optional[Feature] = None
    truth_f: Optional[Truth] = None
    segsnr: Optional[Segsnr] = None


@dataclass
class ImpulseResponseData:
    name: str
    sample_rate: int
    data: AudioT

    @property
    def length(self) -> int:
        return len(self.data)


ImpulseResponseFiles: TypeAlias = list[str]


@dataclass(frozen=True)
class SpectralMask(DataClassSonusAIMixin):
    f_max_width: int
    f_num: int
    t_max_width: int
    t_num: int
    t_max_percent: int


SpectralMasks: TypeAlias = list[SpectralMask]


@dataclass
class Target(DataClassSonusAIMixin):
    file_id: Optional[int] = None
    augmentation: Optional[Augmentation] = None
    gain: Optional[float] = None


Targets = list[Target]


@dataclass
class Noise(DataClassSonusAIMixin):
    file_id: Optional[int] = None
    augmentation: Optional[Augmentation] = None
    offset: Optional[int] = None


@dataclass
class Mixture(DataClassSonusAIMixin):
    name: Optional[str] = None
    noise: Optional[Noise] = None
    noise_snr_gain: Optional[float] = None
    samples: Optional[int] = None
    snr: Optional[UniversalSNR] = None
    spectral_mask_id: Optional[int] = None
    spectral_mask_seed: Optional[int] = None
    target_snr_gain: Optional[float] = None
    targets: Optional[Targets] = None

    @property
    def noise_id(self) -> int:
        return self.noise.file_id

    @property
    def target_id(self) -> list[int]:
        return [target.file_id for target in self.targets]


Mixtures = list[Mixture]


@dataclass(frozen=True)
class TransformConfig:
    N: int
    R: int
    bin_start: int
    bin_end: int
    ttype: str


@dataclass(frozen=True)
class FeatureGeneratorConfig:
    feature_mode: str
    num_classes: int
    truth_mutex: bool


@dataclass(frozen=True)
class FeatureGeneratorInfo:
    decimation: int
    stride: int
    step: int
    num_bands: int
    ft_config: TransformConfig
    eft_config: TransformConfig
    it_config: TransformConfig


@dataclass
class MixtureDatabaseConfig(DataClassSonusAIMixin):
    asr_manifest: list[str] = field(default_factory=list)
    class_balancing: Optional[bool] = False
    class_labels: Optional[list[str]] = None
    class_weights_threshold: Optional[list[float]] = None
    feature: Optional[str] = None
    impulse_response_files: Optional[ImpulseResponseFiles] = None
    mixtures: Optional[Mixtures] = None
    noise_mix_mode: Optional[str] = 'exhaustive'
    noise_files: Optional[NoiseFiles] = None
    num_classes: Optional[int] = None
    spectral_masks: Optional[SpectralMasks] = None
    target_files: Optional[TargetFiles] = None
    truth_mutex: Optional[bool] = None
    truth_reduction_function: Optional[str] = None



"""sonusai genmixdb

usage: genmixdb [-hvmfsd] LOC

options:
   -h, --help
   -v, --verbose    Be verbose.
   -m, --mix        Save mixture data. [default: False].
   -f, --ft         Save feature/truth_f data. [default: False].
   -s, --segsnr     Save segsnr data. [default: False].
   -d, --dryrun     Perform a dry run showing the processed config. [default: False].

Create mixture database data for training and evaluation. Optionally, also create mixture audio and feature/truth data.

genmixdb creates a database of training and evaluation feature and truth data generation information. It allows the
choice of audio neural-network feature types that are supported by the Aaware real-time front-end and truth data that is
synchronized frame-by-frame with the feature data.

Here are some examples:

#### Adding target data
Suppose you have an audio file which is an example, or target, of what you want to recognize or detect. Of course, for
training a NN you also need truth data for that file (also called parameters/labels/classes). If you don't already have
it, genmixdb can create truth using a variety of generation functions on each frame of the feature data. You can also
select different feature types. Here's an example:

genmixdb target_gfr32ts2

where target_gfr32ts2 contains config.yml with the following inside:
---
feature: gfr32ts2

targets:
  - name: data/target.wav

target_augmentations:
  - normalize: -3.5
...

The mixture database is written to a SQLite file (mixdb.db) in the same directory that contains the config.yml file.

#### Target data mix with noise and augmentation

genmixdb mix_gfr32ts2.yml

where mix_gfr32ts2.yml contains:
---
feature: gfr32ts2

targets:
  - name: data/target.wav

target_augmentations:
  - normalize: -3.5
    pitch: [-3, 0, 3]
    tempo: [0.8, 1, 1.2]

noises:
  - name: data/noise.wav

noise_augmentations:
  - normalize: -3.5

snrs:
  - 20
...

In this example a time-domain mixture is created and feature data is calculated as specified by 'feature: gfr32ts2'.
Various feature types are available which vary in spectral and temporal resolution (4 ms or higher), and other feature
algorithm parameters. The total feature size, dimension, and #frames for mixture is reported in the log file (the log
file name is genmixdb.log).

Truth (parameters/labels/classes) can be automatically created per feature output frame based on a variety of truth
generation functions. By default, these are included with the feature data in a single HDF5 output file. By default,
truth generation is turned on with default settings (see truth section) and a single class, i.e., detecting a single
type of sound. The truth format is a single float per class representing the probability of activity/presence, and
multi-class truth is possible by specifying the number of classes and either a scalar index or a vector of indices in
which to put the truth result. For example, 'num_class: 3' and 'truth_index: 2' adds a 1x3 vector to the feature data
with truth put in index 2 (others would be 0) for data/target.wav being an audio clip from sound type of class 2.

The mixture is created with potential data augmentation functions in the following way:
1. apply noise augmentation rule
2. apply target augmentation rule to each target in the mixture (multiple targets may be used in mixup)
3. adjust noise and target gains for specified SNR
4. add augmented noise to augmented target(s)

Note: If an impulse response is part of the target augmentation, truth generation is performed on the targets before
applying the IRs. In this way, the truth is not impacted by the IR.

The mixture length is the length of the longest target in the mixture, and the noise signal is repeated if it is
shorter, or trimmed if longer.

#### Target and noise using path lists

Target and noise audio is specified as a list containing text files, audio files, and file globs. Text files are
processed with items on each line where each item can be a text file, an audio file, or a file glob. Each item will be
searched for audio files which can be WAV, MP3, FLAC, AIFF, or OGG format with any sample rate, bit depth, or channel
count. All audio files will be converted to 16 kHz, float32, single channel (only the first channel is used) format
before processing.

For example,

genmixdb dog-bark.yml

where dog-bark.yml contains:
---
targets:
  - name: slib/dog-outside/*.wav
  - name: slib/dog-inside/*.wav

will find all .wav files in the specified directories and process them as targets.

"""
from dataclasses import dataclass

from sonusai import logger
from sonusai.mixture import Mixture
from sonusai.mixture import MixtureDatabase


@dataclass
class MPGlobal:
    mixdb: MixtureDatabase = None
    save_mix: bool = None
    save_ft: bool = None
    save_segsnr: bool = None


MP_GLOBAL = MPGlobal()


def genmixdb(location: str,
             save_mix: bool = False,
             save_ft: bool = False,
             save_segsnr: bool = False,
             logging: bool = True,
             show_progress: bool = False,
             test: bool = False) -> MixtureDatabase:
    from random import seed

    import yaml
    from tqdm import tqdm

    from sonusai import SonusAIError
    from sonusai import logger
    from sonusai.mixture import AugmentationRule
    from sonusai.mixture import MixtureDatabase
    from sonusai.mixture import SAMPLE_BYTES
    from sonusai.mixture import SAMPLE_RATE
    from sonusai.mixture import balance_targets
    from sonusai.mixture import generate_mixtures
    from sonusai.mixture import get_all_snrs_from_config
    from sonusai.mixture import get_augmentation_rules
    from sonusai.mixture import get_augmented_targets
    from sonusai.mixture import get_impulse_response_files
    from sonusai.mixture import get_mixups
    from sonusai.mixture import get_noise_files
    from sonusai.mixture import get_target_augmentations_for_mixup
    from sonusai.mixture import get_target_files
    from sonusai.mixture import initialize_db
    from sonusai.mixture import load_config
    from sonusai.mixture import log_duration_and_sizes
    from sonusai.mixture import populate_asr_manifest_table
    from sonusai.mixture import populate_class_label_table
    from sonusai.mixture import populate_class_weights_threshold_table
    from sonusai.mixture import populate_impulse_response_file_table
    from sonusai.mixture import populate_mixture_table
    from sonusai.mixture import populate_noise_file_table
    from sonusai.mixture import populate_spectral_mask_table
    from sonusai.mixture import populate_target_file_table
    from sonusai.mixture import populate_top_table
    from sonusai.mixture import update_mixid_width
    from sonusai.utils import dataclass_from_dict
    from sonusai.utils import human_readable_size
    from sonusai.utils import pp_tqdm_imap
    from sonusai.utils import seconds_to_hms

    config = load_config(location)
    initialize_db(location=location, test=test)

    mixdb = MixtureDatabase(location=location, test=test)

    populate_top_table(location, config, test)
    populate_asr_manifest_table(location, config, test)
    populate_class_label_table(location, config, test)
    populate_class_weights_threshold_table(location, config, test)
    populate_spectral_mask_table(location, config, test)

    seed(config['seed'])

    if logging:
        logger.debug(f'Seed: {config["seed"]}')
        logger.debug('Configuration:')
        logger.debug(yaml.dump(config))

    if logging:
        logger.info('Collecting targets')

    target_files = get_target_files(config, show_progress=show_progress)

    if len(target_files) == 0:
        raise SonusAIError('Canceled due to no targets')

    populate_target_file_table(location, target_files, test)

    if logging:
        logger.debug('List of targets:')
        logger.debug(yaml.dump([target.name for target in mixdb.target_files], default_flow_style=False))
        logger.debug('')

    if logging:
        logger.info('Collecting noises')

    noise_files = get_noise_files(config, show_progress=show_progress)

    populate_noise_file_table(location, noise_files, test)

    if logging:
        logger.debug('List of noises:')
        logger.debug(yaml.dump([noise.name for noise in mixdb.noise_files], default_flow_style=False))
        logger.debug('')

    if logging:
        logger.info('Collecting impulse responses')

    impulse_response_files = get_impulse_response_files(config, show_progress=show_progress)

    populate_impulse_response_file_table(location, impulse_response_files, test)

    if logging:
        logger.debug('List of impulse responses:')
        logger.debug(
            yaml.dump([impulse_response for impulse_response in mixdb.impulse_response_files],
                      default_flow_style=False))
        logger.debug('')

    if logging:
        logger.info('Collecting target augmentations')

    target_augmentations = get_augmentation_rules(rules=config['target_augmentations'],
                                                  num_ir=mixdb.num_impulse_response_files)
    mixups = get_mixups(target_augmentations)

    if logging:
        for mixup in mixups:
            logger.debug(f'Expanded list of target augmentation rules for mixup of {mixup}:')
            for target_augmentation in get_target_augmentations_for_mixup(target_augmentations, mixup):
                ta_dict = target_augmentation.to_dict()
                del ta_dict['mixup']
                logger.debug(f'- {ta_dict}')
            logger.debug('')

    if logging:
        logger.info('Collecting noise augmentations')

    noise_augmentations = get_augmentation_rules(rules=config['noise_augmentations'],
                                                 num_ir=mixdb.num_impulse_response_files)

    if logging:
        logger.debug('Expanded list of noise augmentations:')
        for noise_augmentation in noise_augmentations:
            na_dict = noise_augmentation.to_dict()
            del na_dict['mixup']
            logger.debug(f'- {na_dict}')
        logger.debug('')

    if logging:
        logger.debug(f'SNRs: {config["snrs"]}\n')
        logger.debug(f'Random SNRs: {config["random_snrs"]}\n')
        logger.debug(f'Noise mix mode: {mixdb.noise_mix_mode}\n')
        logger.debug(f'Spectral masks:')
        for spectral_mask in mixdb.spectral_masks:
            logger.debug(f'- {spectral_mask}')
        logger.debug('')

    if mixdb.truth_mutex and any(mixup > 1 for mixup in mixups):
        raise SonusAIError(f'Mutex truth mode is not compatible with mixup')

    if logging:
        logger.info('Collecting augmented targets')

    augmented_targets = get_augmented_targets(target_files, target_augmentations, mixups)

    if config['class_balancing']:
        class_balancing_augmentation = dataclass_from_dict(AugmentationRule, config['class_balancing_augmentation'])
        augmented_targets, target_augmentations = balance_targets(
            augmented_targets=augmented_targets,
            targets=target_files,
            target_augmentations=target_augmentations,
            class_balancing_augmentation=class_balancing_augmentation,
            num_classes=mixdb.num_classes,
            truth_mutex=mixdb.truth_mutex,
            num_ir=mixdb.num_impulse_response_files,
            mixups=mixups)

    target_audio_samples = sum([targets.samples for targets in mixdb.target_files])
    target_audio_duration = target_audio_samples / SAMPLE_RATE
    noise_audio_duration = sum([noises.duration for noises in mixdb.noise_files])
    noise_audio_samples = noise_audio_duration * SAMPLE_RATE

    if logging:
        logger.info('')
        logger.info(f'Target audio: {mixdb.num_target_files} files, '
                    f'{human_readable_size(target_audio_samples * SAMPLE_BYTES, 1)}, '
                    f'{seconds_to_hms(seconds=target_audio_duration)}')
        logger.info(f'Noise audio: {mixdb.num_noise_files} files, '
                    f'{human_readable_size(noise_audio_samples * SAMPLE_BYTES, 1)}, '
                    f'{seconds_to_hms(seconds=noise_audio_duration)}')

    used_noise_files, used_noise_samples, mixtures = generate_mixtures(
        noise_mix_mode=mixdb.noise_mix_mode,
        augmented_targets=augmented_targets,
        target_files=target_files,
        target_augmentations=target_augmentations,
        noise_files=noise_files,
        noise_augmentations=noise_augmentations,
        spectral_masks=mixdb.spectral_masks,
        all_snrs=get_all_snrs_from_config(config),
        mixups=mixups,
        num_classes=mixdb.num_classes,
        truth_mutex=mixdb.truth_mutex,
        feature_step_samples=mixdb.feature_step_samples,
        num_ir=mixdb.num_impulse_response_files)

    num_mixtures = len(mixtures)
    update_mixid_width(location, num_mixtures, test)

    if logging:
        logger.info('')
        logger.info(f'Found {num_mixtures:,} mixtures to process')

    total_duration = float(sum([mixture.samples for mixture in mixtures])) / SAMPLE_RATE

    if logging:
        log_duration_and_sizes(total_duration=total_duration,
                               num_classes=mixdb.num_classes,
                               feature_step_samples=mixdb.feature_step_samples,
                               num_bands=mixdb.fg_num_bands,
                               stride=mixdb.fg_stride,
                               desc='Estimated')
        logger.info(f'Feature shape:        '
                    f'{mixdb.fg_stride} x {mixdb.fg_num_bands} '
                    f'({mixdb.fg_stride * mixdb.fg_num_bands} total params)')
        logger.info(f'Feature samples:      {mixdb.feature_samples} samples ({mixdb.feature_ms} ms)')
        logger.info(f'Feature step samples: {mixdb.feature_step_samples} samples ({mixdb.feature_step_ms} ms)')
        logger.info('')

    # Fill in the details
    if logging:
        logger.info('Generating mixtures')
    progress = tqdm(total=num_mixtures, disable=not show_progress)
    mixtures = pp_tqdm_imap(_process_mixture, mixtures,
                            progress=progress,
                            initializer=_initializer,
                            initargs=(location, save_mix, save_ft, save_segsnr, test))
    progress.close()

    populate_mixture_table(location, mixtures, test)

    total_noise_files = len(noise_files)

    total_samples = mixdb.total_samples()
    total_duration = float(total_samples / SAMPLE_RATE)

    noise_files_percent = (float(used_noise_files) / float(total_noise_files)) * 100
    noise_samples_percent = (float(used_noise_samples) / float(noise_audio_samples)) * 100

    if logging:
        log_duration_and_sizes(total_duration=total_duration,
                               num_classes=mixdb.num_classes,
                               feature_step_samples=mixdb.feature_step_samples,
                               num_bands=mixdb.fg_num_bands,
                               stride=mixdb.fg_stride,
                               desc='Actual')
        logger.info('')
        logger.info(f'Used {noise_files_percent:,.0f}% of noise files')
        logger.info(f'Used {noise_samples_percent:,.0f}% of noise audio')
        logger.info('')

    if not test:
        mixdb = MixtureDatabase(location)
        mixdb.save()
        if logging:
            logger.info(f'Wrote mixture database to {location}')

    return mixdb


def _initializer(location: str, save_mix: bool, save_ft: bool, save_segsnr: bool, test: bool) -> None:
    MP_GLOBAL.mixdb = MixtureDatabase(location, test)
    MP_GLOBAL.save_mix = save_mix
    MP_GLOBAL.save_ft = save_ft
    MP_GLOBAL.save_segsnr = save_segsnr


def _process_mixture(mixture: Mixture) -> Mixture:
    from typing import Any

    from sonusai.mixture import get_ft
    from sonusai.mixture import get_segsnr
    from sonusai.mixture import get_truth_t
    from sonusai.mixture import update_mixture
    from sonusai.mixture import write_mixture_data
    from sonusai.mixture import write_mixture_metadata

    with_data = MP_GLOBAL.save_mix or MP_GLOBAL.save_ft
    mixdb = MP_GLOBAL.mixdb

    mixture, genmix_data = update_mixture(mixdb, mixture, with_data)

    if with_data:
        write_data: list[tuple[str, Any]] = []

        if MP_GLOBAL.save_mix:
            write_data.append(('targets', genmix_data.targets))
            write_data.append(('noise', genmix_data.noise))
            write_data.append(('mixture', genmix_data.mixture))

        if MP_GLOBAL.save_ft:
            truth_t = get_truth_t(mixdb=mixdb,
                                  mixture=mixture,
                                  targets_audio=genmix_data.targets,
                                  noise_audio=genmix_data.noise,
                                  mixture_audio=genmix_data.mixture)
            feature, truth_f = get_ft(mixdb=mixdb,
                                      mixture=mixture,
                                      mixture_audio=genmix_data.mixture,
                                      truth_t=truth_t)
            write_data.append(('feature', feature))
            write_data.append(('truth_f', truth_f))

            if MP_GLOBAL.save_segsnr:
                segsnr = get_segsnr(mixdb=mixdb,
                                    mixture=mixture,
                                    target_audio=genmix_data.target,
                                    noise=genmix_data.noise)
                write_data.append(('segsnr', segsnr))

        write_mixture_data(mixdb, mixture, write_data)
        write_mixture_metadata(mixdb, mixture)

    return mixture


def main() -> None:
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    import time
    from os import makedirs
    from os import remove
    from os.path import exists
    from os.path import isdir
    from os.path import join

    import yaml

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import logger
    from sonusai import update_console_handler
    from sonusai.mixture import load_config
    from sonusai.utils import seconds_to_hms

    verbose = args['--verbose']
    save_mix = args['--mix']
    save_ft = args['--ft']
    save_segsnr = args['--segsnr']
    dryrun = args['--dryrun']
    location = args['LOC']

    start_time = time.monotonic()

    if exists(location) and not isdir(location):
        remove(location)

    makedirs(location, exist_ok=True)

    create_file_handler(join(location, 'genmixdb.log'))
    update_console_handler(verbose)
    initial_log_messages('genmixdb')

    if dryrun:
        config = load_config(location)
        logger.info('Dryrun configuration:')
        logger.info(yaml.dump(config))
        return

    logger.info(f'Creating mixture database for {location}')
    logger.info('')

    genmixdb(location=location,
             save_mix=save_mix,
             save_ft=save_ft,
             save_segsnr=save_segsnr,
             show_progress=True)

    end_time = time.monotonic()
    logger.info(f'Completed in {seconds_to_hms(seconds=end_time - start_time)}')
    logger.info('')


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)



from sonusai.mixture.datatypes import Truth
from sonusai.mixture.truth_functions.data import Data


def phoneme(_data: Data) -> Truth:
    """Read in .txt transcript and run a Python function to generate text grid data
(indicating which phonemes are active). Then generate truth based on this data and put
in the correct classes based on the index in the config.
    """
    from sonusai import SonusAIError

    raise SonusAIError('Truth function phoneme is not supported yet')



from dataclasses import dataclass
from typing import Any
from typing import Optional

from sonusai.mixture import AudioT


@dataclass(frozen=True)
class ASRResult:
    text: str
    confidence: Optional[float] = None
    lang: Optional[str] = None
    lang_prob: Optional[float] = None
    duration: Optional[float] = None
    num_segments: Optional[int] = None
    asr_cpu_time: Optional[float] = None


def calc_asr(audio: AudioT | str,
             engine: Optional[str] = 'deepgram',
             whisper_model: Optional[Any] = None,
             whisper_model_name: Optional[str] = 'tiny',
             device: Optional[str] = 'cpu',
             cpu_threads: Optional[int] = 1,
             compute_type: Optional[str] = 'int8',
             beam_size: Optional[int] = 5) -> ASRResult:
    """Run ASR on audio

    :param audio: Numpy array of audio samples or location of an audio file
    :param engine: Type of ASR engine to use
    :param whisper_model: A preloaded Whisper ASR model
    :param whisper_model_name: Name of Whisper ASR model to use if none was provided
    :param device: the device to put the ASR model into
    :param cpu_threads: int specifying threads to use when device is cpu
           note: must be 1 if this func is run in parallel
    :param compute_type: the precision of ASR model to use
    :param beam_size: int specifying beam_size to use
    :return: ASRResult object containing text and confidence
    """
    from copy import copy

    import numpy as np

    from sonusai import SonusAIError
    from sonusai.mixture import read_audio
    from sonusai.utils import asr_functions
    from sonusai.utils.asr_functions.data import Data

    if not isinstance(audio, np.ndarray):
        audio = copy(read_audio(audio))

    data = Data(audio, whisper_model, whisper_model_name, device, cpu_threads, compute_type, beam_size)

    try:
        return getattr(asr_functions, engine)(data)
    except AttributeError:
        raise SonusAIError(f'Unsupported ASR function: {engine}')



from sonusai.mixture.datatypes import ImpulseResponseFiles
from sonusai.mixture.datatypes import NoiseFiles
from sonusai.mixture.datatypes import SpectralMasks
from sonusai.mixture.datatypes import TargetFiles


def raw_load_config(name: str) -> dict:
    """Load YAML config file

    :param name: File name
    :return: Dictionary of config data
    """
    import yaml

    with open(file=name, mode='r') as f:
        config = yaml.safe_load(f)

    return config


def get_default_config() -> dict:
    """Load default SonusAI config

    :return: Dictionary of default config data
    """
    from sonusai import SonusAIError
    from .constants import DEFAULT_CONFIG

    try:
        return raw_load_config(DEFAULT_CONFIG)
    except Exception as e:
        raise SonusAIError(f'Error loading default config: {e}')


def load_config(name: str) -> dict:
    """Load SonusAI default config and update with given location (performing SonusAI variable substitution)

    :param name: Directory containing mixture database
    :return: Dictionary of config data
    """
    from os.path import join

    return update_config_from_file(name=join(name, 'config.yml'), config=get_default_config())


def update_config_from_file(name: str, config: dict) -> dict:
    """Update the given config with the config in the YAML file

    :param name: File name
    :param config: Config dictionary to update
    :return: Updated config dictionary
    """
    from copy import deepcopy

    from sonusai import SonusAIError
    from .constants import REQUIRED_CONFIGS
    from .constants import VALID_CONFIGS
    from .constants import VALID_NOISE_MIX_MODES

    updated_config = deepcopy(config)

    try:
        new_config = raw_load_config(name)
    except Exception as e:
        raise SonusAIError(f'Error loading config from {name}: {e}')

    # Check for unrecognized keys
    for key in new_config:
        if key not in VALID_CONFIGS:
            nice_list = '\n'.join([f'  {item}' for item in VALID_CONFIGS])
            raise SonusAIError(f'Invalid config parameter in {name}: {key}.\n'
                               f'Valid config parameters are:\n{nice_list}')

    # Use default config as base and overwrite with given config keys as found
    for key in updated_config:
        if key in new_config:
            if key not in ['truth_settings']:
                updated_config[key] = new_config[key]

    # Handle 'truth_settings' special case
    if 'truth_settings' in new_config:
        updated_config['truth_settings'] = deepcopy(new_config['truth_settings'])

    if not isinstance(updated_config['truth_settings'], list):
        updated_config['truth_settings'] = [updated_config['truth_settings']]

    default = deepcopy(config['truth_settings'])
    if not isinstance(default, list):
        default = [default]

    updated_config['truth_settings'] = update_truth_settings(updated_config['truth_settings'], default)

    # Check for required keys
    for key in REQUIRED_CONFIGS:
        if key not in updated_config:
            raise SonusAIError(f'Missing required config in {name}: {key}')

    # Check for non-empty spectral masks
    if len(updated_config['spectral_masks']) == 0:
        updated_config['spectral_masks'] = config['spectral_masks']

    # Check for valid noise_mix_mode
    if updated_config['noise_mix_mode'] not in VALID_NOISE_MIX_MODES:
        nice_list = '\n'.join([f'  {item}' for item in VALID_NOISE_MIX_MODES])
        raise SonusAIError(f'Invalid noise_mix_mode in {name}.\n'
                           f'Valid noise mix modes are:\n{nice_list}')

    return updated_config


def update_truth_settings(given: list[dict] | dict, default: list[dict] = None) -> list[dict]:
    """Update missing fields in given 'truth_settings' with default values

    :param given: The dictionary of given truth settings
    :param default: The dictionary of default truth settings
    :return: Updated dictionary of truth settings
    """
    from copy import deepcopy

    from sonusai import SonusAIError
    from .constants import VALID_TRUTH_SETTINGS

    if isinstance(given, list):
        truth_settings = deepcopy(given)
    else:
        truth_settings = [deepcopy(given)]

    if default is not None and len(truth_settings) != len(default):
        raise SonusAIError(f'Length of given does not match default')

    for n in range(len(truth_settings)):
        for key in truth_settings[n]:
            if key not in VALID_TRUTH_SETTINGS:
                nice_list = '\n'.join([f'  {item}' for item in VALID_TRUTH_SETTINGS])
                raise SonusAIError(f'Invalid truth_settings: {key}.\nValid truth_settings are:\n{nice_list}')

        for key in VALID_TRUTH_SETTINGS:
            if key not in truth_settings[n]:
                if default is not None and key in default[n]:
                    truth_settings[n][key] = default[n][key]
                else:
                    raise SonusAIError(f'Missing required truth_settings: {key}')

    for truth_setting in truth_settings:
        if not isinstance(truth_setting['index'], list):
            truth_setting['index'] = [truth_setting['index']]

    return truth_settings


def get_hierarchical_config_files(root: str, leaf: str) -> list[str]:
    """Get a hierarchical list of config files in the given leaf of the given root

    :param root: Root of the hierarchy
    :param leaf: Leaf under the root
    :return: List of config files found in the hierarchy
    """
    import os
    from pathlib import Path

    from sonusai import SonusAIError

    config_file = 'config.yml'

    root_path = Path(os.path.abspath(root))
    if not root_path.is_dir():
        raise SonusAIError(f'Given root, {root_path}, is not a directory.')

    leaf_path = Path(os.path.abspath(leaf))
    if not leaf_path.is_dir():
        raise SonusAIError(f'Given leaf, {leaf_path}, is not a directory.')

    common = os.path.commonpath((root_path, leaf_path))
    if os.path.normpath(common) != os.path.normpath(root_path):
        raise SonusAIError(f'Given leaf, {leaf_path}, is not in the hierarchy of the given root, {root_path}')

    top_config_file = os.path.join(root_path, config_file)
    if not Path(top_config_file).is_file():
        raise SonusAIError(f'Could not find {top_config_file}')

    current = leaf_path
    config_files = []
    while current != root_path:
        local_config_file = Path(os.path.join(current, config_file))
        if local_config_file.is_file():
            config_files.append(str(local_config_file))
        current = current.parent

    config_files.append(top_config_file)
    return list(reversed(config_files))


def update_config_from_hierarchy(root: str, leaf: str, config: dict) -> dict:
    """Update the given config using the hierarchical config files in the given leaf of the given root

    :param root: Root of the hierarchy
    :param leaf: Leaf under the root
    :param config: Config to update
    :return: Updated config
    """
    from copy import deepcopy

    new_config = deepcopy(config)
    config_files = get_hierarchical_config_files(root=root, leaf=leaf)
    for config_file in config_files:
        new_config = update_config_from_file(name=config_file, config=new_config)

    return new_config


def get_max_class(num_classes: int, truth_mutex: bool) -> int:
    """Get the maximum class index

    :param num_classes: Number of classes
    :param truth_mutex: Truth is mutex mode
    :return: Highest class index
    """
    max_class = num_classes
    if truth_mutex:
        max_class -= 1
    return max_class


def get_target_files(config: dict, show_progress: bool = False) -> TargetFiles:
    """Get the list of target files from a config

    :param config: Config dictionary
    :param show_progress: Show progress bar
    :return: List of target files
    """
    from itertools import chain

    from tqdm import tqdm

    from sonusai import SonusAIError
    from sonusai.utils import dataclass_from_dict
    from sonusai.utils import pp_tqdm_imap
    from .datatypes import TargetFiles

    truth_settings = config.get('truth_settings', list())
    level_type = config.get('target_level_type', None)
    target_files = list(chain.from_iterable([append_target_files(entry=entry,
                                                                 truth_settings=truth_settings,
                                                                 level_type=level_type)
                                             for entry in config['targets']]))

    progress = tqdm(total=len(target_files), disable=not show_progress)
    target_files = pp_tqdm_imap(_get_num_samples, target_files, progress=progress)
    progress.close()

    max_class = get_max_class(config['num_classes'], config['truth_mode'] == 'mutex')

    for target_file in target_files:
        target_file['truth_settings'] = update_truth_settings(target_file['truth_settings'], config['truth_settings'])

        for truth_setting in target_file['truth_settings']:
            if any(idx > max_class for idx in truth_setting['index']):
                raise SonusAIError('invalid truth index')

    return dataclass_from_dict(TargetFiles, target_files)


def append_target_files(entry: dict | str,
                        truth_settings: list[dict],
                        level_type: str | None,
                        tokens: dict = None) -> list[dict]:
    """Process target files list and append as needed

    :param entry: Target file entry to append to the list
    :param truth_settings: Truth settings
    :param level_type: Target level type
    :param tokens: Tokens used for variable expansion
    :return: List of target files
    """
    from glob import glob
    from os import listdir
    from os.path import dirname
    from os.path import isabs
    from os.path import isdir
    from os.path import join
    from os.path import splitext

    from sonusai import SonusAIError
    from .audio import validate_input_file
    from .tokenized_shell_vars import tokenized_expand
    from .tokenized_shell_vars import tokenized_replace

    if tokens is None:
        tokens = {}

    if isinstance(entry, dict):
        if 'name' in entry:
            in_name = entry['name']
        else:
            raise SonusAIError('Target list contained record without name')

        if 'truth_settings' in entry:
            truth_settings = entry['truth_settings']
        if 'target_level_type' in entry:
            level_type = entry['target_level_type']
    else:
        in_name = entry

    in_name, new_tokens = tokenized_expand(in_name)
    tokens.update(new_tokens)
    names = sorted(glob(in_name))
    if not names:
        raise SonusAIError(f'Could not find {in_name}. Make sure path exists')

    target_files: list[dict] = []
    for name in names:
        ext = splitext(name)[1].lower()
        dir_name = dirname(name)
        if isdir(name):
            for file in listdir(name):
                child = file
                if not isabs(child):
                    child = join(dir_name, child)
                target_files.extend(append_target_files(entry=child,
                                                        truth_settings=truth_settings,
                                                        level_type=level_type,
                                                        tokens=tokens))
        else:
            try:
                if ext == '.txt':
                    with open(file=name, mode='r') as txt_file:
                        for line in txt_file:
                            # strip comments
                            child = line.partition('#')[0]
                            child = child.rstrip()
                            if child:
                                child, new_tokens = tokenized_expand(child)
                                tokens.update(new_tokens)
                                if not isabs(child):
                                    child = join(dir_name, child)
                                target_files.extend(append_target_files(entry=child,
                                                                        truth_settings=truth_settings,
                                                                        level_type=level_type,
                                                                        tokens=tokens))
                elif ext == '.yml':
                    try:
                        yml_config = raw_load_config(name)

                        if 'targets' in yml_config:
                            for record in yml_config['targets']:
                                target_files.extend(append_target_files(entry=record,
                                                                        truth_settings=truth_settings,
                                                                        level_type=level_type,
                                                                        tokens=tokens))
                    except Exception as e:
                        raise SonusAIError(f'Error processing {name}: {e}')
                else:
                    validate_input_file(name)
                    target_file: dict = {
                        'expanded_name': name,
                        'name':          tokenized_replace(name, tokens),
                    }
                    if len(truth_settings) > 0:
                        target_file['truth_settings'] = truth_settings
                        for truth_setting in target_file['truth_settings']:
                            if 'function' in truth_setting and truth_setting['function'] == 'file':
                                truth_setting['config']['file'] = splitext(target_file['name'])[0] + '.h5'
                    if level_type is not None:
                        target_file['level_type'] = level_type
                    target_files.append(target_file)
            except SonusAIError:
                raise
            except Exception as e:
                raise SonusAIError(f'Error processing {name}: {e}')

    return target_files


def get_noise_files(config: dict, show_progress: bool = False) -> NoiseFiles:
    """Get the list of noise files from a config

    :param config: Config dictionary
    :param show_progress: Show progress bar
    :return: List of noise file
    """
    from itertools import chain

    from tqdm import tqdm

    from sonusai.utils import dataclass_from_dict
    from sonusai.utils import pp_tqdm_imap
    from .datatypes import NoiseFiles

    noise_files = list(chain.from_iterable([append_noise_files(entry=entry) for entry in config['noises']]))

    progress = tqdm(total=len(noise_files), disable=not show_progress)
    noise_files = pp_tqdm_imap(_get_num_samples, noise_files, progress=progress)
    progress.close()

    return dataclass_from_dict(NoiseFiles, noise_files)


def append_noise_files(entry: dict | str, tokens: dict = None) -> list[dict]:
    """Process noise files list and append as needed

    :param entry: Noise file entry to append to the list
    :param tokens: Tokens used for variable expansion
    :return: List of noise files
    """
    from glob import glob
    from os import listdir
    from os.path import dirname
    from os.path import isabs
    from os.path import isdir
    from os.path import join
    from os.path import splitext

    from sonusai import SonusAIError
    from .audio import validate_input_file
    from .tokenized_shell_vars import tokenized_expand
    from .tokenized_shell_vars import tokenized_replace

    if tokens is None:
        tokens = {}

    if isinstance(entry, dict):
        if 'name' in entry:
            in_name = entry['name']
        else:
            raise SonusAIError('Noise list contained record without name')
    else:
        in_name = entry

    in_name, new_tokens = tokenized_expand(in_name)
    tokens.update(new_tokens)
    names = sorted(glob(in_name))
    if not names:
        raise SonusAIError(f'Could not find {in_name}. Make sure path exists')

    noise_files: list[dict] = []
    for name in names:
        ext = splitext(name)[1].lower()
        dir_name = dirname(name)
        if isdir(name):
            for file in listdir(name):
                child = file
                if not isabs(child):
                    child = join(dir_name, child)
                noise_files.extend(append_noise_files(entry=child, tokens=tokens))
        else:
            try:
                if ext == '.txt':
                    with open(file=name, mode='r') as txt_file:
                        for line in txt_file:
                            # strip comments
                            child = line.partition('#')[0]
                            child = child.rstrip()
                            if child:
                                child, new_tokens = tokenized_expand(child)
                                tokens.update(new_tokens)
                                if not isabs(child):
                                    child = join(dir_name, child)
                                noise_files.extend(append_noise_files(entry=child, tokens=tokens))
                elif ext == '.yml':
                    try:
                        yml_config = raw_load_config(name)

                        if 'noises' in yml_config:
                            for record in yml_config['noises']:
                                noise_files.extend(append_noise_files(entry=record, tokens=tokens))
                    except Exception as e:
                        raise SonusAIError(f'Error processing {name}: {e}')
                else:
                    validate_input_file(name)
                    noise_file: dict = {
                        'expanded_name': name,
                        'name':          tokenized_replace(name, tokens),
                    }
                    noise_files.append(noise_file)
            except SonusAIError:
                raise
            except Exception as e:
                raise SonusAIError(f'Error processing {name}: {e}')

    return noise_files


def get_impulse_response_files(config: dict, show_progress: bool = False) -> ImpulseResponseFiles:
    """Get the list of impulse response files from a config

    :param config: Config dictionary
    :param show_progress: Show progress bar
    :return: List of impulse response files
    """
    from itertools import chain

    return list(
        chain.from_iterable([append_impulse_response_files(entry=entry) for entry in config['impulse_responses']]))


def append_impulse_response_files(entry: str, tokens: dict = None) -> list[str]:
    """Process impulse response files list and append as needed

    :param entry: Impulse response file entry to append to the list
    :param tokens: Tokens used for variable expansion
    :return: List of impulse response files
    """
    from glob import glob
    from os import listdir
    from os.path import dirname
    from os.path import isabs
    from os.path import isdir
    from os.path import join
    from os.path import splitext

    from sonusai import SonusAIError
    from .audio import validate_input_file
    from .tokenized_shell_vars import tokenized_expand
    from .tokenized_shell_vars import tokenized_replace

    if tokens is None:
        tokens = {}

    in_name, new_tokens = tokenized_expand(entry)
    tokens.update(new_tokens)
    names = sorted(glob(in_name))
    if not names:
        raise SonusAIError(f'Could not find {in_name}. Make sure path exists')

    impulse_response_files: list[str] = []
    for name in names:
        ext = splitext(name)[1].lower()
        dir_name = dirname(name)
        if isdir(name):
            for file in listdir(name):
                child = file
                if not isabs(child):
                    child = join(dir_name, child)
                impulse_response_files.extend(append_impulse_response_files(entry=child, tokens=tokens))
        else:
            try:
                if ext == '.txt':
                    with open(file=name, mode='r') as txt_file:
                        for line in txt_file:
                            # strip comments
                            child = line.partition('#')[0]
                            child = child.rstrip()
                            if child:
                                child, new_tokens = tokenized_expand(child)
                                tokens.update(new_tokens)
                                if not isabs(child):
                                    child = join(dir_name, child)
                                impulse_response_files.extend(append_impulse_response_files(entry=child, tokens=tokens))
                elif ext == '.yml':
                    try:
                        yml_config = raw_load_config(name)

                        if 'impulse_responses' in yml_config:
                            for record in yml_config['impulse_responses']:
                                impulse_response_files.extend(
                                    append_impulse_response_files(entry=record, tokens=tokens))
                    except Exception as e:
                        raise SonusAIError(f'Error processing {name}: {e}')
                else:
                    validate_input_file(name)
                    impulse_response_files.append(tokenized_replace(name, tokens))
            except SonusAIError:
                raise
            except Exception as e:
                raise SonusAIError(f'Error processing {name}: {e}')

    return impulse_response_files


def get_spectral_masks(config: dict) -> SpectralMasks:
    """Get the list of spectral masks from a config

    :param config: Config dictionary
    :return: List of spectral masks
    """
    from sonusai import SonusAIError
    from sonusai.utils import dataclass_from_dict
    from .datatypes import SpectralMasks

    try:
        return dataclass_from_dict(SpectralMasks, config['spectral_masks'])
    except Exception as e:
        raise SonusAIError(f'Error in spectral_masks: {e}')


def _get_num_samples(entry: dict) -> dict:
    from .audio import get_num_samples

    entry['samples'] = get_num_samples(entry['expanded_name'])
    del entry['expanded_name']
    return entry



from sonusai.mixture.datatypes import AudioT
from sonusai.mixture.datatypes import Augmentation
from sonusai.mixture.datatypes import ImpulseResponseData


def apply_augmentation(audio: AudioT,
                       augmentation: Augmentation,
                       frame_length: int = 1) -> AudioT:
    """Apply augmentations to audio data using torchaudio.sox_effects

    :param audio: Audio
    :param augmentation: Augmentation
    :param frame_length: Pad resulting audio to be a multiple of this
    :return: Augmented audio
    """
    import numpy as np
    import torch
    import torchaudio

    from sonusai import SonusAIError
    from .augmentation import pad_audio_to_frame
    from .constants import SAMPLE_RATE

    effects: list[list[str]] = []

    # TODO
    #  Always normalize and remove normalize from list of available augmentations
    #  Normalize to globally set level (should this be a global config parameter,
    #  or hard-coded into the script?)
    if augmentation.normalize is not None:
        effects.append(['norm', str(augmentation.normalize)])

    if augmentation.gain is not None:
        effects.append(['gain', str(augmentation.gain)])

    if augmentation.pitch is not None:
        effects.append(['pitch', str(augmentation.pitch)])
        effects.append(['rate', str(SAMPLE_RATE)])

    if augmentation.tempo is not None:
        effects.append(['tempo', '-s', str(augmentation.tempo)])

    if augmentation.eq1 is not None:
        effects.append(['equalizer', *[str(item) for item in augmentation.eq1]])

    if augmentation.eq2 is not None:
        effects.append(['equalizer', *[str(item) for item in augmentation.eq2]])

    if augmentation.eq3 is not None:
        effects.append(['equalizer', *[str(item) for item in augmentation.eq3]])

    if augmentation.lpf is not None:
        effects.append(['lowpass', '-2', str(augmentation.lpf), '0.707'])

    if effects:
        if audio.ndim == 1:
            audio = np.reshape(audio, (1, audio.shape[0]))
        out = torch.tensor(audio)

        try:
            out, _ = torchaudio.sox_effects.apply_effects_tensor(out,
                                                                 sample_rate=SAMPLE_RATE,
                                                                 effects=effects)
        except Exception as e:
            raise SonusAIError(f'Error applying {augmentation}: {e}')

        audio_out = np.squeeze(np.array(out))
    else:
        audio_out = audio

    # make sure length is multiple of frame_length
    return pad_audio_to_frame(audio=audio_out, frame_length=frame_length)


def apply_impulse_response(audio: AudioT, ir: ImpulseResponseData) -> AudioT:
    """Apply impulse response to audio data using torchaudio.fftconvolve

    :param audio: Audio
    :param ir: Impulse response data
    :return: Augmented audio
    """
    import numpy as np
    import torch
    import torchaudio

    from sonusai.utils import linear_to_db
    from .constants import SAMPLE_RATE

    # Early exit if no ir or if all audio is zero
    if ir is None or not audio.any():
        return audio

    # Get current maximum level in dB
    max_db = linear_to_db(max(abs(audio)))

    # Convert audio to IR sample rate
    audio_in = torch.reshape(torch.tensor(audio), (1, len(audio)))
    audio_out, sr = torchaudio.sox_effects.apply_effects_tensor(audio_in,
                                                                sample_rate=SAMPLE_RATE,
                                                                effects=[['rate', str(ir.sample_rate)]])

    # Apply IR and convert back to global sample rate
    rir = torch.reshape(torch.tensor(ir.data), (1, len(ir.data)))
    audio_out = torchaudio.functional.fftconvolve(audio_out, rir)
    audio_out, sr = torchaudio.sox_effects.apply_effects_tensor(audio_out,
                                                                sample_rate=ir.sample_rate,
                                                                effects=[['rate', str(SAMPLE_RATE)]])

    # Reset level to previous max value
    audio_out, sr = torchaudio.sox_effects.apply_effects_tensor(audio_out,
                                                                sample_rate=SAMPLE_RATE,
                                                                effects=[['norm', str(max_db)]])

    return np.squeeze(np.array(audio_out[:, :len(audio)]))



def convert_string_to_number(string: str) -> float | int | str:
    try:
        result = float(string)
        if result == int(result):
            return int(result)
        return result
    except ValueError:
        return string



from sonusai.utils import ASRResult
from sonusai.utils.asr_functions.data import Data


def whisper(data: Data) -> ASRResult:
    from whisper import load_model

    whisper_model = data.whisper_model
    if whisper_model is None:
        whisper_model = load_model(data.whisper_model_name, device=data.device)

    return ASRResult(text=whisper_model.transcribe(data.audio, fp16=False)['text'])


"""
Whisper results:
{
  'text': ' The birch canoe slid on the smooth planks.',
  'segments': [
    {
      'id': 0,
      'seek': 0,
      'start': 0.0,
      'end': 2.4,
      'text': ' The birch canoe slid on the smooth planks.',
      'tokens': [
        50363,
        383,
        35122,
        354,
        47434,
        27803,
        319,
        262,
        7209,
        1410,
        591,
        13,
        50483
      ],
      'temperature': 0.0,
      'avg_logprob': -0.4188103675842285,
      'compression_ratio': 0.8571428571428571,
      'no_speech_prob': 0.003438911633566022
    }
  ],
  'language': 'en'
}
"""



from sonusai.mixture.datatypes import AudioT
from sonusai.mixture.datatypes import ImpulseResponseData


def _raw_read(name: str) -> tuple[AudioT, int]:
    import numpy as np
    import soundfile
    from pydub import AudioSegment

    from sonusai import SonusAIError
    from .tokenized_shell_vars import tokenized_expand

    expanded_name, _ = tokenized_expand(name)

    try:
        if expanded_name.endswith('.mp3'):
            sound = AudioSegment.from_mp3(expanded_name)
            raw = np.array(sound.get_array_of_samples()).astype(np.float32).reshape((-1, sound.channels))
            raw = raw / 2 ** (sound.sample_width * 8 - 1)
            sample_rate = sound.frame_rate
        else:
            raw, sample_rate = soundfile.read(expanded_name, always_2d=True, dtype='float32')
    except Exception as e:
        if name != expanded_name:
            raise SonusAIError(f'Error reading {name} (expanded: {expanded_name}): {e}')
        else:
            raise SonusAIError(f'Error reading {name}: {e}')

    return np.squeeze(raw[:, 0]), sample_rate


def read_ir(name: str) -> ImpulseResponseData:
    """Read impulse response data using soundfile

    :param name: File name
    :return: ImpulseResponseData object
    """
    import numpy as np

    from .datatypes import ImpulseResponseData

    # Read impulse response data from audio file
    out, sample_rate = _raw_read(name)
    offset = np.argmax(out)
    out = out[offset:]
    out = out / np.linalg.norm(out)

    return ImpulseResponseData(name=name, sample_rate=sample_rate, data=out)


def read_audio(name: str) -> AudioT:
    """Read audio data from a file using soundfile

    :param name: File name
    :return: Array of time domain audio data
    """
    import librosa

    from .constants import SAMPLE_RATE

    out, sample_rate = _raw_read(name)
    out = librosa.resample(out,
                           orig_sr=sample_rate,
                           target_sr=SAMPLE_RATE,
                           res_type='soxr_hq')

    return out


def get_num_samples(name: str) -> int:
    """Get the number of samples resampled to the SonusAI sample rate in the given file

    :param name: File name
    :return: number of samples in resampled audio
    """
    import math

    import soundfile
    from pydub import AudioSegment

    from .constants import SAMPLE_RATE
    from .tokenized_shell_vars import tokenized_expand

    expanded_name, _ = tokenized_expand(name)

    if expanded_name.endswith('.mp3'):
        sound = AudioSegment.from_mp3(expanded_name)
        samples = sound.frame_count()
        sample_rate = sound.frame_rate
    else:
        info = soundfile.info(name)
        samples = info.frames
        sample_rate = info.samplerate
    return math.ceil(SAMPLE_RATE * samples / sample_rate)



from sonusai.utils.asr_manifest_functions import PathInfo
from sonusai.utils.asr_manifest_functions import TranscriptData


def collect_vctk_noisy_speech_transcripts(paths: list[str] | str) -> TranscriptData:
    from glob import iglob
    from os import listdir
    from os.path import abspath
    from os.path import basename
    from os.path import join
    from os.path import split
    from os.path import splitext

    from sonusai import SonusAIError

    entries: TranscriptData = {}
    if not isinstance(paths, list):
        paths = [paths]

    for p in paths:
        abs_p = abspath(p)
        head, tail = split(abs_p)

        dirs = listdir(head)
        tail = tail.replace('wav', 'txt')

        location = None
        for d in dirs:
            if tail.endswith(d):
                location = join(head, d, '*.txt')
                break
        if location is None:
            raise SonusAIError(f'Could not find VCTK Noisy Speech transcript data for {p}')

        for file in iglob(pathname=location, recursive=True):
            with open(file, encoding='utf-8') as f:
                lines = f.readlines()
                if len(lines) != 1:
                    raise SonusAIError(f'Ill-formed VCTK Noisy Speech transcript file: {file}')

                name = join(abs_p, splitext(basename(file))[0])
                text = lines[0].lower().strip()

                if name in entries:
                    raise SonusAIError(f'{name} already exists in transcript data')
                entries[name] = text.lower().strip()

    return entries


def get_vctk_noisy_speech_manifest_entry(entry: PathInfo, transcript_data: TranscriptData) -> dict:
    from os.path import splitext
    from subprocess import check_output

    from sonusai import SonusAIError

    name = splitext(entry.abs_path)[0]
    duration = float(check_output(f'soxi -D {entry.abs_path}', shell=True))
    if name not in transcript_data.keys():
        raise SonusAIError(f'Could not find {name} in transcript data')

    return {
        'audio_filepath': entry.audio_filepath,
        'text':           transcript_data[name],
        'duration':       duration,
    }



from typing import Generator
from typing import Optional
from typing import Set


def expand_braces(text: str, seen: Optional[Set[str]] = None) -> Generator[str, None, None]:
    """Brace-expansion pre-processing for glob.

    Expand all the braces, then run glob on each of the results.
    (Brace-expansion turns one string into a list of strings.)
    https://stackoverflow.com/questions/22996645/brace-expansion-in-python-glob
    """
    import itertools
    import re

    if seen is None:
        seen = set()

    spans = [m.span() for m in re.finditer(r'\{[^{}]*}', text)][::-1]
    alts = [text[start + 1: stop - 1].split(',') for start, stop in spans]

    if len(spans) == 0:
        if text not in seen:
            yield text
        seen.add(text)
    else:
        for combo in itertools.product(*alts):
            replaced = list(text)
            for (start, stop), replacement in zip(spans, combo):
                replaced[start:stop] = replacement
            yield from expand_braces(''.join(replaced), seen)


def braced_glob(pathname: str, recursive: bool = False) -> list[str]:
    from glob import glob

    result = []
    for expanded_path in expand_braces(pathname):
        result.extend(glob(expanded_path, recursive=recursive))

    return result


def braced_iglob(pathname: str, recursive: bool = False) -> Generator[str, None, None]:
    from glob import iglob

    for expanded_path in expand_braces(pathname):
        yield from iglob(expanded_path, recursive=recursive)



"""sonusai calc_metric_spenh

usage: calc_metric_spenh [-hvtpws] [-i MIXID] [-e WER] [-m WMNAME] PLOC TLOC

options:
    -h, --help
    -v, --verbose               Be verbose.
    -i MIXID, --mixid MIXID     Mixture ID(s) to process, can be range like 0:maxmix+1 [default: *].
    -t, --truth-est-mode        Calculate extraction and metrics using truth (instead of prediction).
    -p, --plot                  Enable PDF plots file generation per mixture.
    -w, --wav                   Generate WAV files per mixture.
    -s, --summary               Enable summary files generation.
    -e WER, --wer-method WER    Word-Error-Rate method: deepgram, google, aixplain_whisper
                                or whisper (locally run) [default: none]
    -m WMNAME, --whisper-model  Whisper model name used in aixplain_whisper and whisper WER methods.
                                [default: tiny]

Calculate speech enhancement metrics of prediction data in PLOC using SonusAI mixture data
in TLOC as truth/label reference. Metric and extraction data files are written into PLOC.

PLOC  directory containing prediction data in .h5 files created from truth/label mixture data in TLOC
TLOC  directory with SonusAI mixture database of truth/label mixture data

For whisper WER methods, the possible models used in local processing (WER = whisper) are:
    {tiny.en,tiny,base.en,base,small.en,small,medium.en,medium,large-v1,large-v2,large}
but note most are very computationally demanding and can overwhelm/hang a local system.

Outputs the following to PLOC (where id is mixd number 0:num_mixtures):
    <id>_metric_spenh.txt

    If --plot:
        <id>_metric_spenh.pdf

    If --wav:
        <id>_target.wav
        <id>_target_est.wav
        <id>_noise.wav
        <id>_noise_est.wav
        <id>_mixture.wav

        If --truth-est-mode:
            <id>_target_truth_est.wav
            <id>_noise_truth_est.wav

    If --summary:
        metric_spenh_targetf_summary.txt
        metric_spenh_targetf_summary.csv
        metric_spenh_targetf_list.csv
        metric_spenh_targetf_estats_list.csv

        If --truth-est-mode:
            metric_spenh_targetf_truth_list.csv
            metric_spenh_targetf_estats_truth_list.csv

TBD
Metric and extraction data are written into prediction location PLOC as separate files per mixture.

    -d PLOC, --ploc PLOC        Location of SonusAI predict data.

Inputs:

"""
from dataclasses import dataclass
from typing import Optional

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from sonusai import logger
from sonusai.mixture import AudioF
from sonusai.mixture import AudioT
from sonusai.mixture import Feature
from sonusai.mixture import MixtureDatabase
from sonusai.mixture import Predict

matplotlib.use('SVG')


@dataclass
class MPGlobal:
    mixdb: MixtureDatabase = None
    predict_location: str = None
    predwav_mode: bool = None
    truth_est_mode: bool = None
    enable_plot: bool = None
    enable_wav: bool = None
    wer_method: str = None
    whisper_model: str = None


MP_GLOBAL = MPGlobal()


def power_compress(spec):
    mag = np.abs(spec)
    phase = np.angle(spec)
    mag = mag ** 0.3
    real_compress = mag * np.cos(phase)
    imag_compress = mag * np.sin(phase)
    return real_compress + 1j * imag_compress


def power_uncompress(spec):
    mag = np.abs(spec)
    phase = np.angle(spec)
    mag = mag ** (1. / 0.3)
    real_uncompress = mag * np.cos(phase)
    imag_uncompress = mag * np.sin(phase)
    return real_uncompress + 1j * imag_uncompress


def snr(clean_speech, processed_speech, sample_rate):
    # Check the length of the clean and processed speech. Must be the same.
    clean_length = len(clean_speech)
    processed_length = len(processed_speech)
    if clean_length != processed_length:
        raise ValueError('Both Speech Files must be same length.')

    overall_snr = 10 * np.log10(np.sum(np.square(clean_speech)) / np.sum(np.square(clean_speech - processed_speech)))

    # Global Variables
    winlength = round(30 * sample_rate / 1000)  # window length in samples
    skiprate = int(np.floor(winlength / 4))  # window skip in samples
    MIN_SNR = -10  # minimum SNR in dB
    MAX_SNR = 35  # maximum SNR in dB

    # For each frame of input speech, calculate the Segmental SNR
    num_frames = int(clean_length / skiprate - (winlength / skiprate))  # number of frames
    start = 0  # starting sample
    window = 0.5 * (1 - np.cos(2 * np.pi * np.arange(1, winlength + 1) / (winlength + 1)))

    segmental_snr = np.empty(num_frames)
    EPS = np.spacing(1)
    for frame_count in range(num_frames):
        # (1) Get the Frames for the test and reference speech. Multiply by Hanning Window.
        clean_frame = clean_speech[start:start + winlength]
        processed_frame = processed_speech[start:start + winlength]
        clean_frame = np.multiply(clean_frame, window)
        processed_frame = np.multiply(processed_frame, window)

        # (2) Compute the Segmental SNR
        signal_energy = np.sum(np.square(clean_frame))
        noise_energy = np.sum(np.square(clean_frame - processed_frame))
        segmental_snr[frame_count] = 10 * np.log10(signal_energy / (noise_energy + EPS) + EPS)
        segmental_snr[frame_count] = max(segmental_snr[frame_count], MIN_SNR)
        segmental_snr[frame_count] = min(segmental_snr[frame_count], MAX_SNR)

        start = start + skiprate

    return overall_snr, segmental_snr


def lpcoeff(speech_frame, model_order):
    # (1) Compute Autocorrelation Lags
    winlength = np.size(speech_frame)
    R = np.empty(model_order + 1)
    E = np.empty(model_order + 1)
    for k in range(model_order + 1):
        R[k] = np.dot(speech_frame[0:winlength - k], speech_frame[k: winlength])

    # (2) Levinson-Durbin
    a = np.ones(model_order)
    a_past = np.empty(model_order)
    rcoeff = np.empty(model_order)
    E[0] = R[0]
    for i in range(model_order):
        a_past[0: i] = a[0: i]
        sum_term = np.dot(a_past[0: i], R[i:0:-1])
        rcoeff[i] = (R[i + 1] - sum_term) / E[i]
        a[i] = rcoeff[i]
        if i == 0:
            a[0: i] = a_past[0: i] - np.multiply(a_past[i - 1:-1:-1], rcoeff[i])
        else:
            a[0: i] = a_past[0: i] - np.multiply(a_past[i - 1::-1], rcoeff[i])
        E[i + 1] = (1 - rcoeff[i] * rcoeff[i]) * E[i]
    acorr = R
    refcoeff = rcoeff
    lpparams = np.concatenate((np.array([1]), -a))
    return acorr, refcoeff, lpparams


def llr(clean_speech, processed_speech, sample_rate):
    from scipy.linalg import toeplitz

    # Check the length of the clean and processed speech.  Must be the same.
    clean_length = np.size(clean_speech)
    processed_length = np.size(processed_speech)
    if clean_length != processed_length:
        raise ValueError('Both Speech Files must be same length.')

    # Global Variables
    winlength = (np.round(30 * sample_rate / 1000)).astype(int)  # window length in samples
    skiprate = (np.floor(winlength / 4)).astype(int)  # window skip in samples
    if sample_rate < 10000:
        P = 10  # LPC Analysis Order
    else:
        P = 16  # this could vary depending on sampling frequency.

    # For each frame of input speech, calculate the Log Likelihood Ratio
    num_frames = int((clean_length - winlength) / skiprate)  # number of frames
    start = 0  # starting sample
    window = 0.5 * (1 - np.cos(2 * np.pi * np.arange(1, winlength + 1) / (winlength + 1)))

    distortion = np.empty(num_frames)
    for frame_count in range(num_frames):
        # (1) Get the Frames for the test and reference speech. Multiply by Hanning Window.
        clean_frame = clean_speech[start: start + winlength]
        processed_frame = processed_speech[start: start + winlength]
        clean_frame = np.multiply(clean_frame, window)
        processed_frame = np.multiply(processed_frame, window)

        # (2) Get the autocorrelation lags and LPC parameters used to compute the LLR measure.
        R_clean, Ref_clean, A_clean = lpcoeff(clean_frame, P)
        R_processed, Ref_processed, A_processed = lpcoeff(processed_frame, P)

        # (3) Compute the LLR measure
        numerator = np.dot(np.matmul(A_processed, toeplitz(R_clean)), A_processed)
        denominator = np.dot(np.matmul(A_clean, toeplitz(R_clean)), A_clean)
        distortion[frame_count] = np.log(numerator / denominator)
        start = start + skiprate
    return distortion


def wss(clean_speech, processed_speech, sample_rate):
    from scipy.fftpack import fft

    # Check the length of the clean and processed speech, which must be the same.
    clean_length = np.size(clean_speech)
    processed_length = np.size(processed_speech)
    if clean_length != processed_length:
        raise ValueError('Files must have same length.')

    # Global variables
    winlength = (np.round(30 * sample_rate / 1000)).astype(int)  # window length in samples
    skiprate = (np.floor(np.divide(winlength, 4))).astype(int)  # window skip in samples
    max_freq = (np.divide(sample_rate, 2)).astype(int)  # maximum bandwidth
    num_crit = 25  # number of critical bands

    USE_FFT_SPECTRUM = 1  # defaults to 10th order LP spectrum
    n_fft = (np.power(2, np.ceil(np.log2(2 * winlength)))).astype(int)
    n_fftby2 = (np.multiply(0.5, n_fft)).astype(int)  # FFT size/2
    Kmax = 20.0  # value suggested by Klatt, pg 1280
    Klocmax = 1.0  # value suggested by Klatt, pg 1280

    # Critical Band Filter Definitions (Center Frequency and Bandwidths in Hz)
    cent_freq = np.array([50.0000, 120.000, 190.000, 260.000, 330.000, 400.000, 470.000,
                          540.000, 617.372, 703.378, 798.717, 904.128, 1020.38, 1148.30,
                          1288.72, 1442.54, 1610.70, 1794.16, 1993.93, 2211.08, 2446.71,
                          2701.97, 2978.04, 3276.17, 3597.63])
    bandwidth = np.array([70.0000, 70.0000, 70.0000, 70.0000, 70.0000, 70.0000, 70.0000,
                          77.3724, 86.0056, 95.3398, 105.411, 116.256, 127.914, 140.423,
                          153.823, 168.154, 183.457, 199.776, 217.153, 235.631, 255.255,
                          276.072, 298.126, 321.465, 346.136])

    bw_min = bandwidth[0]  # minimum critical bandwidth

    # Set up the critical band filters.
    # Note here that Gaussianly shaped filters are used.
    # Also, the sum of the filter weights are equivalent for each critical band filter.
    # Filter less than -30 dB and set to zero.
    min_factor = np.exp(-30.0 / (2.0 * 2.303))  # -30 dB point of filter
    crit_filter = np.empty((num_crit, n_fftby2))
    for i in range(num_crit):
        f0 = (cent_freq[i] / max_freq) * n_fftby2
        bw = (bandwidth[i] / max_freq) * n_fftby2
        norm_factor = np.log(bw_min) - np.log(bandwidth[i])
        j = np.arange(n_fftby2)
        crit_filter[i, :] = np.exp(-11 * np.square(np.divide(j - np.floor(f0), bw)) + norm_factor)
        cond = np.greater(crit_filter[i, :], min_factor)
        crit_filter[i, :] = np.where(cond, crit_filter[i, :], 0)
    # For each frame of input speech, calculate the Weighted Spectral Slope Measure
    num_frames = int(clean_length / skiprate - (winlength / skiprate))  # number of frames
    start = 0  # starting sample
    window = 0.5 * (1 - np.cos(2 * np.pi * np.arange(1, winlength + 1) / (winlength + 1)))

    distortion = np.empty(num_frames)
    for frame_count in range(num_frames):
        # (1) Get the Frames for the test and reference speech. Multiply by Hanning Window.
        clean_frame = clean_speech[start: start + winlength] / 32768
        processed_frame = processed_speech[start: start + winlength] / 32768
        clean_frame = np.multiply(clean_frame, window)
        processed_frame = np.multiply(processed_frame, window)
        # (2) Compute the Power Spectrum of Clean and Processed
        # if USE_FFT_SPECTRUM:
        clean_spec = np.square(np.abs(fft(clean_frame, n_fft)))
        processed_spec = np.square(np.abs(fft(processed_frame, n_fft)))

        # (3) Compute Filterbank Output Energies (in dB scale)
        clean_energy = np.matmul(crit_filter, clean_spec[0:n_fftby2])
        processed_energy = np.matmul(crit_filter, processed_spec[0:n_fftby2])

        clean_energy = 10 * np.log10(np.maximum(clean_energy, 1E-10))
        processed_energy = 10 * np.log10(np.maximum(processed_energy, 1E-10))

        # (4) Compute Spectral Slope (dB[i+1]-dB[i])
        clean_slope = clean_energy[1:num_crit] - clean_energy[0: num_crit - 1]
        processed_slope = processed_energy[1:num_crit] - processed_energy[0: num_crit - 1]

        # (5) Find the nearest peak locations in the spectra to each critical band.
        #     If the slope is negative, we search to the left. If positive, we search to the right.
        clean_loc_peak = np.empty(num_crit - 1)
        processed_loc_peak = np.empty(num_crit - 1)

        for i in range(num_crit - 1):
            # find the peaks in the clean speech signal
            if clean_slope[i] > 0:  # search to the right
                n = i
                while (n < num_crit - 1) and (clean_slope[n] > 0):
                    n = n + 1
                clean_loc_peak[i] = clean_energy[n - 1]
            else:  # search to the left
                n = i
                while (n >= 0) and (clean_slope[n] <= 0):
                    n = n - 1
                clean_loc_peak[i] = clean_energy[n + 1]

            # find the peaks in the processed speech signal
            if processed_slope[i] > 0:  # search to the right
                n = i
                while (n < num_crit - 1) and (processed_slope[n] > 0):
                    n = n + 1
                processed_loc_peak[i] = processed_energy[n - 1]
            else:  # search to the left
                n = i
                while (n >= 0) and (processed_slope[n] <= 0):
                    n = n - 1
                processed_loc_peak[i] = processed_energy[n + 1]

        # (6) Compute the WSS Measure for this frame. This includes determination of the weighting function.
        dBMax_clean = np.max(clean_energy)
        dBMax_processed = np.max(processed_energy)
        '''
        The weights are calculated by averaging individual weighting factors from the clean and processed frame.
        These weights W_clean and W_processed should range from 0 to 1 and place more emphasis on spectral peaks
        and less emphasis on slope differences in spectral valleys.
        This procedure is described on page 1280 of Klatt's 1982 ICASSP paper.
        '''
        Wmax_clean = np.divide(Kmax, Kmax + dBMax_clean - clean_energy[0: num_crit - 1])
        Wlocmax_clean = np.divide(Klocmax, Klocmax + clean_loc_peak - clean_energy[0: num_crit - 1])
        W_clean = np.multiply(Wmax_clean, Wlocmax_clean)

        Wmax_processed = np.divide(Kmax, Kmax + dBMax_processed - processed_energy[0: num_crit - 1])
        Wlocmax_processed = np.divide(Klocmax, Klocmax + processed_loc_peak - processed_energy[0: num_crit - 1])
        W_processed = np.multiply(Wmax_processed, Wlocmax_processed)

        W = np.divide(np.add(W_clean, W_processed), 2.0)
        slope_diff = np.subtract(clean_slope, processed_slope)[0: num_crit - 1]
        distortion[frame_count] = np.dot(W, np.square(slope_diff)) / np.sum(W)
        # this normalization is not part of Klatt's paper, but helps to normalize the measure.
        # Here we scale the measure by the sum of the weights.
        start = start + skiprate
    return distortion


def calc_speech_metrics(hypothesis: np.ndarray,
                        reference: np.ndarray) -> tuple[float, int, int, int, float]:
    """
    Calculate speech metrics pesq_mos, CSIG, CBAK, COVL, segSNR.  These are all related and thus included
    in one function. Reference: matlab script "compute_metrics.m".

    Usage:
        pesq, csig, cbak, covl, ssnr = compute_metrics(hypothesis, reference, Fs, path)
        reference: clean audio as array
        hypothesis: enhanced audio as array
        Audio must have sampling rate = 16000 Hz.

    Example call:
        pesq_output, csig_output, cbak_output, covl_output, ssnr_output = \
                calc_speech_metrics(predicted_audio, target_audio)
    """
    from sonusai.metrics import calc_pesq

    Fs = 16000

    # compute the WSS measure
    wss_dist_vec = wss(reference, hypothesis, Fs)
    wss_dist_vec = np.sort(wss_dist_vec)
    alpha = 0.95  # value from CMGAN ref implementation
    wss_dist = np.mean(wss_dist_vec[0: round(np.size(wss_dist_vec) * alpha)])

    # compute the LLR measure
    llr_dist = llr(reference, hypothesis, Fs)
    ll_rs = np.sort(llr_dist)
    llr_len = round(np.size(llr_dist) * alpha)
    llr_mean = np.mean(ll_rs[0: llr_len])

    # compute the SNRseg
    snr_dist, segsnr_dist = snr(reference, hypothesis, Fs)
    snr_mean = snr_dist
    segSNR = np.mean(segsnr_dist)

    # compute the pesq (use Sonusai wrapper, only fs=16k, mode=wb support)
    pesq_mos = calc_pesq(hypothesis=hypothesis, reference=reference)
    # pesq_mos = pesq(sampling_rate1, data1, data2, 'wb')

    # now compute the composite measures
    CSIG = 3.093 - 1.029 * llr_mean + 0.603 * pesq_mos - 0.009 * wss_dist
    CSIG = max(1, CSIG)
    CSIG = min(5, CSIG)  # limit values to [1, 5]
    CBAK = 1.634 + 0.478 * pesq_mos - 0.007 * wss_dist + 0.063 * segSNR
    CBAK = max(1, CBAK)
    CBAK = min(5, CBAK)  # limit values to [1, 5]
    COVL = 1.594 + 0.805 * pesq_mos - 0.512 * llr_mean - 0.007 * wss_dist
    COVL = max(1, COVL)
    COVL = min(5, COVL)  # limit values to [1, 5]

    return pesq_mos, CSIG, CBAK, COVL, segSNR


def mean_square_error(hypothesis: np.ndarray,
                      reference: np.ndarray,
                      squared: bool = False) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Calculate root-mean-square error or mean square error

    :param hypothesis: [frames, bins]
    :param reference: [frames, bins]
    :param squared: calculate mean square rather than root-mean-square
    :return: mean, mean per bin, mean per frame
    """
    sq_err = np.square(reference - hypothesis)

    # mean over frames for value per bin
    err_b = np.mean(sq_err, axis=0)
    # mean over bins for value per frame
    err_f = np.mean(sq_err, axis=1)
    # mean over all
    err = np.mean(sq_err)

    if not squared:
        err_b = np.sqrt(err_b)
        err_f = np.sqrt(err_f)
        err = np.sqrt(err)

    return err, err_b, err_f


def mean_abs_percentage_error(hypothesis: np.ndarray,
                              reference: np.ndarray) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Calculate mean abs percentage error

    If inputs are complex, calculates average: mape(real)/2 + mape(imag)/2

    :param hypothesis: [frames, bins]
    :param reference: [frames, bins]
    :return: mean, mean per bin, mean per frame
    """
    if not np.iscomplexobj(reference) and not np.iscomplexobj(hypothesis):
        abs_err = 100 * np.abs((reference - hypothesis) / (reference + np.finfo(np.float32).eps))
    else:
        reference_r = np.real(reference)
        reference_i = np.imag(reference)
        hypothesis_r = np.real(hypothesis)
        hypothesis_i = np.imag(hypothesis)
        abs_err_r = 100 * np.abs((reference_r - hypothesis_r) / (reference_r + np.finfo(np.float32).eps))
        abs_err_i = 100 * np.abs((reference_i - hypothesis_i) / (reference_i + np.finfo(np.float32).eps))
        abs_err = (abs_err_r / 2) + (abs_err_i / 2)

    # mean over frames for value per bin
    err_b = np.around(np.mean(abs_err, axis=0), 3)
    # mean over bins for value per frame
    err_f = np.around(np.mean(abs_err, axis=1), 3)
    # mean over all
    err = np.around(np.mean(abs_err), 3)

    return err, err_b, err_f


def log_error(reference: np.ndarray, hypothesis: np.ndarray) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Calculate log error

    :param reference: complex or real [frames, bins]
    :param hypothesis: complex or real [frames, bins]
    :return: mean, mean per bin, mean per frame
    """
    reference_sq = np.real(reference * np.conjugate(reference))
    hypothesis_sq = np.real(hypothesis * np.conjugate(hypothesis))
    log_err = abs(10 * np.log10((reference_sq + np.finfo(np.float32).eps) / (hypothesis_sq + np.finfo(np.float32).eps)))
    # log_err = abs(10 * np.log10(reference_sq / (hypothesis_sq + np.finfo(np.float32).eps) + np.finfo(np.float32).eps))

    # mean over frames for value per bin
    err_b = np.around(np.mean(log_err, axis=0), 3)
    # mean over bins for value per frame
    err_f = np.around(np.mean(log_err, axis=1), 3)
    # mean over all
    err = np.around(np.mean(log_err), 3)

    return err, err_b, err_f


def phase_distance(reference: np.ndarray,
                   hypothesis: np.ndarray,
                   eps: float = 1e-9) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Calculate weighted phase distance error (weight normalization over bins per frame)

    :param reference: complex [frames, bins]
    :param hypothesis: complex [frames, bins]
    :param eps: epsilon value
    :return: mean, mean per bin, mean per frame
    """
    ang_diff = np.angle(reference) - np.angle(hypothesis)
    phd_mod = (ang_diff + np.pi) % (2 * np.pi) - np.pi
    rh_angle_diff = phd_mod * 180 / np.pi  # angle diff in deg

    # Use complex divide to intrinsically keep angle diff +/-180 deg, but avoid div by zero (real hyp)
    # hyp_real = np.real(hypothesis)
    # near_zeros = np.real(hyp_real) < eps
    # hyp_real = hyp_real * (np.logical_not(near_zeros))
    # hyp_real = hyp_real + (near_zeros * eps)
    # hypothesis = hyp_real + 1j*np.imag(hypothesis)
    # rh_angle_diff = np.angle(reference / hypothesis) * 180 / np.pi  # angle diff +/-180

    # weighted mean over all (scalar)
    reference_mag = np.abs(reference)
    ref_weight = reference_mag / (np.sum(reference_mag) + eps)  # frames x bins
    err = np.around(np.sum(ref_weight * rh_angle_diff), 3)

    # weighted mean over frames (value per bin)
    err_b = np.zeros(reference.shape[1])
    for bi in range(reference.shape[1]):
        ref_weight = reference_mag[:, bi] / (np.sum(reference_mag[:, bi], axis=0) + eps)
        err_b[bi] = np.around(np.sum(ref_weight * rh_angle_diff[:, bi]), 3)

    # weighted mean over bins (value per frame)
    err_f = np.zeros(reference.shape[0])
    for fi in range(reference.shape[0]):
        ref_weight = reference_mag[fi, :] / (np.sum(reference_mag[fi, :]) + eps)
        err_f[fi] = np.around(np.sum(ref_weight * rh_angle_diff[fi, :]), 3)

    return err, err_b, err_f


def plot_mixpred(mixture: AudioT,
                 mixture_f: AudioF,
                 target: Optional[AudioT] = None,
                 feature: Optional[Feature] = None,
                 predict: Optional[Predict] = None,
                 tp_title: str = '') -> plt.Figure:
    from sonusai.mixture import SAMPLE_RATE

    num_plots = 2
    if feature is not None:
        num_plots += 1
    if predict is not None:
        num_plots += 1

    fig, ax = plt.subplots(num_plots, 1, constrained_layout=True, figsize=(11, 8.5))

    # Plot the waveform
    p = 0
    x_axis = np.arange(len(mixture), dtype=np.float32) / SAMPLE_RATE
    ax[p].plot(x_axis, mixture, label='Mixture', color='mistyrose')
    ax[0].set_ylabel('magnitude', color='tab:blue')
    ax[p].set_xlim(x_axis[0], x_axis[-1])
    # ax[p].set_ylim([-1.025, 1.025])
    if target is not None:  # Plot target time-domain waveform on top of mixture
        ax[0].plot(x_axis, target, label='Target', color='tab:blue')
        # ax[0].tick_params(axis='y', labelcolor=color)
    ax[p].set_title('Waveform')

    # Plot the mixture spectrogram
    p += 1
    ax[p].imshow(np.transpose(mixture_f), aspect='auto', interpolation='nearest', origin='lower')
    ax[p].set_title('Mixture')

    if feature is not None:
        p += 1
        ax[p].imshow(np.transpose(feature), aspect='auto', interpolation='nearest', origin='lower')
        ax[p].set_title('Feature')

    if predict is not None:
        p += 1
        im = ax[p].imshow(np.transpose(predict), aspect='auto', interpolation='nearest', origin='lower')
        ax[p].set_title('Predict ' + tp_title)
        plt.colorbar(im, location='bottom')

    return fig


def plot_pdb_predtruth(predict: np.ndarray,
                       truth_f: Optional[np.ndarray] = None,
                       metric: Optional[np.ndarray] = None,
                       tp_title: str = '') -> plt.Figure:
    """Plot predict and optionally truth and a metric in power db, e.g. applies 10*log10(predict)"""
    num_plots = 2
    if truth_f is not None:
        num_plots += 1

    fig, ax = plt.subplots(num_plots, 1, constrained_layout=True, figsize=(11, 8.5))

    # Plot the predict spectrogram
    p = 0
    tmp = 10 * np.log10(predict.transpose() + np.finfo(np.float32).eps)
    im = ax[p].imshow(tmp, aspect='auto', interpolation='nearest', origin='lower')
    ax[p].set_title('Predict')
    plt.colorbar(im, location='bottom')

    if truth_f is not None:
        p += 1
        tmp = 10 * np.log10(truth_f.transpose() + np.finfo(np.float32).eps)
        im = ax[p].imshow(tmp, aspect='auto', interpolation='nearest', origin='lower')
        ax[p].set_title('Truth')
        plt.colorbar(im, location='bottom')

    # Plot the predict avg, and optionally truth avg and metric lines
    pred_avg = 10 * np.log10(np.mean(predict, axis=-1) + np.finfo(np.float32).eps)
    p += 1
    x_axis = np.arange(len(pred_avg), dtype=np.float32)  # / SAMPLE_RATE
    ax[p].plot(x_axis, pred_avg, color='black', linestyle='dashed', label='Predict mean over freq.')
    ax[p].set_ylabel('mean db', color='black')
    ax[p].set_xlim(x_axis[0], x_axis[-1])
    if truth_f is not None:
        truth_avg = 10 * np.log10(np.mean(truth_f, axis=-1) + np.finfo(np.float32).eps)
        ax[p].plot(x_axis, truth_avg, color='green', linestyle='dashed', label='Truth mean over freq.')

    if metric is not None:  # instantiate 2nd y-axis that shares the same x-axis
        ax2 = ax[p].twinx()
        color2 = 'red'
        ax2.plot(x_axis, metric, color=color2, label='sig distortion (mse db)')
        ax2.set_xlim(x_axis[0], x_axis[-1])
        ax2.set_ylim([0, np.max(metric)])
        ax2.set_ylabel('spectral distortion (mse db)', color=color2)
        ax2.tick_params(axis='y', labelcolor=color2)
        ax[p].set_title('SNR and SNR mse (mean over freq. db)')
    else:
        ax[p].set_title('SNR (mean over freq. db)')
        # ax[0].tick_params(axis='y', labelcolor=color)
    return fig


def plot_epredtruth(predict: np.ndarray,
                    predict_wav: np.ndarray,
                    truth_f: Optional[np.ndarray] = None,
                    truth_wav: Optional[np.ndarray] = None,
                    metric: Optional[np.ndarray] = None,
                    tp_title: str = '') -> plt.Figure:
    """Plot predict spectrogram and waveform and optionally truth and a metric)"""
    num_plots = 2
    if truth_f is not None:
        num_plots += 1
    if metric is not None:
        num_plots += 1

    fig, ax = plt.subplots(num_plots, 1, constrained_layout=True, figsize=(11, 8.5))

    # Plot the predict spectrogram
    p = 0
    im = ax[p].imshow(predict.transpose(), aspect='auto', interpolation='nearest', origin='lower')
    ax[p].set_title('Predict')
    plt.colorbar(im, location='bottom')

    if truth_f is not None:  # plot truth if provided and use same colormap as predict
        p += 1
        ax[p].imshow(truth_f.transpose(), im.cmap, aspect='auto', interpolation='nearest', origin='lower')
        ax[p].set_title('Truth')

    # Plot the predict wav, and optionally truth avg and metric lines
    p += 1
    x_axis = np.arange(len(predict_wav), dtype=np.float32)  # / SAMPLE_RATE
    ax[p].plot(x_axis, predict_wav, color='black', linestyle='dashed', label='Speech Estimate')
    ax[p].set_ylabel('Amplitude', color='black')
    ax[p].set_xlim(x_axis[0], x_axis[-1])
    if truth_wav is not None:
        ntrim = len(truth_wav) - len(predict_wav)
        if ntrim > 0:
            truth_wav = truth_wav[0:-ntrim]
        ax[p].plot(x_axis, truth_wav, color='green', linestyle='dashed', label='True Target')

    # Plot the metric lines
    if metric is not None:
        p += 1
        if metric.ndim > 1:  # if it has multiple dims, plot 1st
            metric1 = metric[:, 0]
        else:
            metric1 = metric  # if single dim, plot it as 1st
        x_axis = np.arange(len(metric1), dtype=np.float32)  # / SAMPLE_RATE
        ax[p].plot(x_axis, metric1, color='red', label='Target LogErr')
        ax[p].set_ylabel('log error db', color='red')
        ax[p].set_xlim(x_axis[0], x_axis[-1])
        ax[p].set_ylim([-0.01, np.max(metric1) + .01])
        if metric.ndim > 1:
            if metric.shape[1] > 1:
                metr2 = metric[:, 1]
                ax2 = ax[p].twinx()
                color2 = 'blue'
                ax2.plot(x_axis, metr2, color=color2, label='phase dist (deg)')
                # ax2.set_ylim([-180.0, +180.0])
                if np.max(metr2) - np.min(metr2) > .1:
                    ax2.set_ylim([np.min(metr2), np.max(metr2)])
                ax2.set_ylabel('phase dist (deg)', color=color2)
                ax2.tick_params(axis='y', labelcolor=color2)
                # ax[p].set_title('SNR and SNR mse (mean over freq. db)')

    return fig


def _process_mixture(mixid: int) -> tuple[pd.DataFrame, pd.DataFrame]:
    from os.path import basename
    from os.path import join
    from os.path import splitext

    import h5py
    from numpy import inf
    from pystoi import stoi

    from sonusai import SonusAIError
    from sonusai import logger
    from sonusai.metrics import calc_pcm
    from sonusai.metrics import calc_wer
    from sonusai.metrics import calc_wsdr
    from sonusai.mixture import forward_transform
    from sonusai.mixture import inverse_transform
    from sonusai.mixture import read_audio
    from sonusai.utils import calc_asr
    from sonusai.utils import float_to_int16
    from sonusai.utils import reshape_outputs
    from sonusai.utils import stack_complex
    from sonusai.utils import unstack_complex
    from sonusai.utils import write_wav

    mixdb = MP_GLOBAL.mixdb
    predict_location = MP_GLOBAL.predict_location
    predwav_mode = MP_GLOBAL.predwav_mode
    truth_est_mode = MP_GLOBAL.truth_est_mode
    enable_plot = MP_GLOBAL.enable_plot
    enable_wav = MP_GLOBAL.enable_wav
    wer_method = MP_GLOBAL.wer_method
    whisper_model = MP_GLOBAL.whisper_model

    # 1)  Read predict data, var predict with shape [BatchSize,Classes] or [BatchSize,Tsteps,Classes]
    output_name = join(predict_location, mixdb.mixture(mixid).name)
    predict = None
    if truth_est_mode:
        # in truth estimation mode we use the truth in place of prediction to see metrics with perfect input
        # don't bother to read prediction, and predict var will get assigned to truth later
        # mark outputs with tru suffix, i.e. 0000_truest_*
        base_name = splitext(output_name)[0] + '_truest'
    else:
        base_name, ext = splitext(output_name)  # base_name used later
        if not predwav_mode:
            try:
                with h5py.File(output_name, 'r') as f:
                    predict = np.array(f['predict'])
            except Exception as e:
                raise SonusAIError(f'Error reading {output_name}: {e}')
            # reshape to always be [frames,classes] where ndim==3 case frames = batch * tsteps
            if predict.ndim > 2:  # TBD generalize to somehow detect if timestep dim exists, some cases > 2 don't have
                # logger.debug(f'Prediction reshape from {predict.shape} to remove timestep dimension.')
                predict, _ = reshape_outputs(predict=predict, truth=None, timesteps=predict.shape[1])
        else:
            base_name, ext = splitext(output_name)
            prfname = join(base_name + '.wav')
            audio = read_audio(prfname)
            predict = forward_transform(audio, mixdb.ft_config)
            if mixdb.feature[0:1] == 'h':
                predict = power_compress(predict)
            predict = stack_complex(predict)

    # 2) Collect true target, noise, mixture data, trim to predict size if needed
    target = mixdb.mixture_target(mixid)
    target_f = mixdb.mixture_target_f(mixid, target=target)
    noise = mixdb.mixture_noise(mixid)
    noise_f = mixdb.mixture_noise_f(mixid, noise=noise)
    mixture = mixdb.mixture_mixture(mixid, target=target, noise=noise)
    mixture_f = mixdb.mixture_mixture_f(mixid, mixture=mixture)
    segsnr_f = mixdb.mixture_segsnr(mixid, target=target, noise=noise)
    segsnr_f[segsnr_f == inf] = 7.944e8  # 99db
    segsnr_f[segsnr_f == -inf] = 1.258e-10  # -99db
    # need to use inv-tf to match #samples & latency shift properties of predict inv tf
    targetfi = inverse_transform(target_f, mixdb.it_config)
    noisefi = inverse_transform(noise_f, mixdb.it_config)
    # mixturefi = mixdb.inverse_transform(mixture_f)

    # gen feature, truth - note feature only used for plots
    # TBD parse truth_f for different formats and also multi-truth
    feature, truth_f = mixdb.mixture_ft(mixid, mixture=mixture)
    truth_type = mixdb.target_file(mixdb.mixture(mixid).targets[0].file_id).truth_settings[0].function
    if truth_type == 'target_mixture_f':
        half = truth_f.shape[-1] // 2
        truth_f = truth_f[..., :half]  # extract target_f only

    if not truth_est_mode:
        if predict.shape[0] < target_f.shape[0]:  # target_f, truth_f, mixture_f, etc. same size
            trimf = target_f.shape[0] - predict.shape[0]
            logger.debug(f'Warning: prediction frames less than mixture, trimming {trimf} frames from all truth.')
            target_f = target_f[0:-trimf, :]
            targetfi, _ = inverse_transform(target_f, mixdb.it_config)
            trimt = target.shape[0] - targetfi.shape[0]
            target = target[0:-trimt]
            noise_f = noise_f[0:-trimf, :]
            noise = noise[0:-trimt]
            mixture_f = mixture_f[0:-trimf, :]
            mixture = mixture[0:-trimt]
            truth_f = truth_f[0:-trimf, :]
        elif predict.shape[0] > target_f.shape[0]:
            raise SonusAIError(
                f'Error: prediction has more frames than true mixture {predict.shape[0]} vs {truth_f.shape[0]}')

    # 3) Extraction - format proper complex and wav estimates and truth (unstack, uncompress, inv tf, etc.)
    if truth_est_mode:
        predict = truth_f  # substitute truth for the prediction (for test/debug)
        predict_complex = unstack_complex(predict)  # unstack
        # if feat has compressed mag and truth does not, compress it
        if mixdb.feature[0:1] == 'h' and mixdb.target_file(1).truth_settings[0].function[0:10] != 'targetcmpr':
            predict_complex = power_compress(predict_complex)  # from uncompressed truth
    else:
        predict_complex = unstack_complex(predict)

    truth_f_complex = unstack_complex(truth_f)
    if mixdb.feature[0:1] == 'h':  # 'hn' or 'ha' or 'hd', etc.:  # if feat has compressed mag
        # estimate noise in uncompressed-mag domain
        noise_est_complex = mixture_f - power_uncompress(predict_complex)
        predict_complex = power_uncompress(predict_complex)  # uncompress if truth is compressed
    else:  # cn, c8, ..
        noise_est_complex = mixture_f - predict_complex

    target_est_wav = inverse_transform(predict_complex, mixdb.it_config)
    noise_est_wav = inverse_transform(noise_est_complex, mixdb.it_config)

    # 4) Metrics
    # Target/Speech logerr - PSD estimation accuracy symmetric mean log-spectral distortion
    lerr_tg, lerr_tg_bin, lerr_tg_frame = log_error(reference=truth_f_complex, hypothesis=predict_complex)
    # Noise logerr - PSD estimation accuracy
    lerr_n, lerr_n_bin, lerr_n_frame = log_error(reference=noise_f, hypothesis=noise_est_complex)
    # PCM loss metric
    ytrue_f = np.concatenate((truth_f_complex[:, np.newaxis, :], noise_f[:, np.newaxis, :]), axis=1)
    ypred_f = np.concatenate((predict_complex[:, np.newaxis, :], noise_est_complex[:, np.newaxis, :]), axis=1)
    pcm, pcm_bin, pcm_frame = calc_pcm(hypothesis=ypred_f, reference=ytrue_f, with_log=True)

    # Phase distance
    phd, phd_bin, phd_frame = phase_distance(hypothesis=predict_complex, reference=truth_f_complex)

    # Noise td logerr
    # lerr_nt, lerr_nt_bin, lerr_nt_frame = log_error(noisefi, noise_truth_est_audio)

    # # SA-SDR (time-domain source-aggragated SDR)
    ytrue = np.concatenate((targetfi[:, np.newaxis], noisefi[:, np.newaxis]), axis=1)
    ypred = np.concatenate((target_est_wav[:, np.newaxis], noise_est_wav[:, np.newaxis]), axis=1)
    # # note: w/o scale is more pessimistic number
    # sa_sdr, _ = calc_sa_sdr(hypothesis=ypred, reference=ytrue)
    target_stoi = stoi(targetfi, target_est_wav, 16000, extended=False)

    wsdr, wsdr_cc, wsdr_cw = calc_wsdr(hypothesis=ypred, reference=ytrue, with_log=True)
    # logger.debug(f'wsdr weight sum for mixid {mixid} = {np.sum(wsdr_cw)}.')
    # logger.debug(f'wsdr cweights = {wsdr_cw}.')
    # logger.debug(f'wsdr ccoefs for mixid {mixid} = {wsdr_cc}.')

    # Speech intelligibility measure - PESQ
    if int(mixdb.mixture(mixid).snr) > -99:
        # len = target_est_wav.shape[0]
        pesq_speech, csig_tg, cbak_tg, covl_tg, sgsnr_tg = calc_speech_metrics(target_est_wav, targetfi)
        pesq_mixture, csig_mx, cbak_mx, covl_mx, sgsnr_mx = calc_speech_metrics(mixture, target)
        # pesq_speech_tst = calc_pesq(hypothesis=target_est_wav, reference=target)
        # pesq_mixture_tst = calc_pesq(hypothesis=mixture, reference=target)
        # pesq improvement
        pesq_impr = pesq_speech - pesq_mixture
        # pesq improvement %
        pesq_impr_pc = pesq_impr / (pesq_mixture + np.finfo(np.float32).eps) * 100
    else:
        pesq_speech = 0
        pesq_mixture = 0
        pesq_impr_pc = np.float32(0)
        csig_mx = 0
        csig_tg = 0
        cbak_mx = 0
        cbak_tg = 0
        covl_mx = 0
        covl_tg = 0

    # Calc WER
    asr_tt = ''
    asr_mx = ''
    asr_tge = ''
    if wer_method == 'none' or mixdb.mixture(mixid).snr == -99:  # noise only, ignore/reset target asr
        wer_mx = float('nan')
        wer_tge = float('nan')
        wer_pi = float('nan')
    else:
        if MP_GLOBAL.mixdb.asr_manifests:
            asr_tt = MP_GLOBAL.mixdb.mixture_asr_data(mixid)[0]  # ignore mixup
        else:
            asr_tt = calc_asr(target, engine=wer_method, whisper_model_name=whisper_model).text  # target truth

        if asr_tt:
            asr_mx = calc_asr(mixture, engine=wer_method, whisper_model=whisper_model).text
            asr_tge = calc_asr(target_est_wav, engine=wer_method, whisper_model=whisper_model).text

            wer_mx = calc_wer(asr_mx, asr_tt).wer * 100  # mixture wer
            wer_tge = calc_wer(asr_tge, asr_tt).wer * 100  # target estimate wer
            if wer_mx == 0.0:
                if wer_tge == 0.0:
                    wer_pi = 0.0
                else:
                    wer_pi = -999.0
            else:
                wer_pi = 100 * (wer_mx - wer_tge) / wer_mx
        else:
            print(f'Warning: mixid {mixid} asr truth is empty, setting to 0% wer')
            wer_mx = float(0)
            wer_tge = float(0)
            wer_pi = float(0)

    # 5) Save per mixture metric results
    # Single row in table of scalar metrics per mixture
    mtable1_col = ['MXSNR', 'MXPESQ', 'PESQ', 'PESQi%', 'MXWER', 'WER', 'WERi%', 'WSDR', 'STOI',
                   'PCM', 'SPLERR', 'NLERR', 'PD', 'MXCSIG', 'CSIG', 'MXCBAK', 'CBAK', 'MXCOVL', 'COVL',
                   'SPFILE', 'NFILE']
    ti = mixdb.mixture(mixid).targets[0].file_id
    ni = mixdb.mixture(mixid).noise.file_id
    metr1 = [mixdb.mixture(mixid).snr, pesq_mixture, pesq_speech, pesq_impr_pc, wer_mx, wer_tge, wer_pi, wsdr,
             target_stoi, pcm, lerr_tg, lerr_n, phd, csig_mx, csig_tg, cbak_mx, cbak_tg, covl_mx, covl_tg,
             basename(mixdb.target_file(ti).name), basename(mixdb.noise_file(ni).name)]
    mtab1 = pd.DataFrame([metr1], columns=mtable1_col, index=[mixid])

    # Stats of per frame estimation metrics
    metr2 = pd.DataFrame({'SSNR':  segsnr_f,
                          'PCM':   pcm_frame,
                          'SLERR': lerr_tg_frame,
                          'NLERR': lerr_n_frame,
                          'SPD':   phd_frame})
    metr2 = metr2.describe()  # Use pandas stat function
    metr2['SSNR'][1:] = metr2['SSNR'][1:].apply(
        lambda x: 10 * np.log10(x + 1.01e-10))  # Change SSNR stats to dB, except count
    # create a single row in multi-column header
    new_labels = pd.MultiIndex.from_product([metr2.columns,
                                             ['Avg', 'Min', 'Med', 'Max', 'Std']],
                                            names=['Metric', 'Stat'])
    dat1row = metr2.loc[['mean', 'min', '50%', 'max', 'std'], :].T.stack().to_numpy().reshape((1, -1))
    mtab2 = pd.DataFrame(dat1row,
                         index=[mixid],
                         columns=new_labels)
    mtab2.insert(0, 'MXSNR', mixdb.mixture(mixid).snr, False)  # add MXSNR as the first metric column

    all_metrics_table_1 = mtab1  # return to be collected by process
    all_metrics_table_2 = mtab2  # return to be collected by process

    metric_name = base_name + '_metric_spenh.txt'
    with open(metric_name, 'w') as f:
        print('Speech enhancement metrics:', file=f)
        print(mtab1.round(2).to_string(float_format=lambda x: "{:.2f}".format(x)), file=f)
        print('', file=f)
        print(f'Extraction statistics over {mixture_f.shape[0]} frames:', file=f)
        print(metr2.round(2).to_string(float_format=lambda x: "{:.2f}".format(x)), file=f)
        print('', file=f)
        print(f'Target path: {mixdb.target_file(ti).name}', file=f)
        print(f'Noise path: {mixdb.noise_file(ni).name}', file=f)
        if wer_method != 'none':
            print(f'WER method: {wer_method} and whisper model (if used):  {whisper_model}', file=f)
            if mixdb.asr_manifests:
                print(f'ASR truth from manifest:  {asr_tt}', file=f)
            else:
                print(f'ASR truth from wer method:  {asr_tt}', file=f)
            print(f'ASR result for mixture:  {asr_mx}', file=f)
            print(f'ASR result for prediction:  {asr_tge}', file=f)

        print(f'Augmentations: {mixdb.mixture(mixid)}', file=f)

    # 7) write wav files
    if enable_wav:
        write_wav(name=base_name + '_mixture.wav', audio=float_to_int16(mixture))
        write_wav(name=base_name + '_target.wav', audio=float_to_int16(target))
        # write_wav(name=base_name + '_targetfi.wav', audio=float_to_int16(targetfi))
        write_wav(name=base_name + '_noise.wav', audio=float_to_int16(noise))
        write_wav(name=base_name + '_target_est.wav', audio=float_to_int16(target_est_wav))
        write_wav(name=base_name + '_noise_est.wav', audio=float_to_int16(noise_est_wav))

        # debug code to test for perfect reconstruction of the extraction method
        # note both 75% olsa-hanns and 50% olsa-hann modes checked to have perfect reconstruction
        # target_r = mixdb.inverse_transform(target_f)
        # noise_r = mixdb.inverse_transform(noise_f)
        # _write_wav(name=base_name + '_target_r.wav', audio=float_to_int16(target_r))
        # _write_wav(name=base_name + '_noise_r.wav', audio=float_to_int16(noise_r)) # chk perfect rec

    # 8) Write out plot file
    if enable_plot:
        from matplotlib.backends.backend_pdf import PdfPages
        plot_fname = base_name + '_metric_spenh.pdf'

        # Reshape feature to eliminate overlap redundancy for easier to understand spectrogram view
        # Original size (frames, stride, num_bands), decimates in stride dimension only if step is > 1
        # Reshape to get frames*decimated_stride, num_bands
        step = int(mixdb.feature_samples / mixdb.feature_step_samples)
        if feature.ndim != 3:
            raise SonusAIError(f'feature does not have 3 dimensions: frames, stride, num_bands')

        # for feature cn*00n**
        feat_sgram = unstack_complex(feature)
        feat_sgram = 20 * np.log10(abs(feat_sgram) + np.finfo(np.float32).eps)
        feat_sgram = feat_sgram[:, -step:, :]  # decimate,  Fx1xB
        feat_sgram = np.reshape(feat_sgram, (feat_sgram.shape[0] * feat_sgram.shape[1], feat_sgram.shape[2]))

        with PdfPages(plot_fname) as pdf:
            # page1 we always have a mixture and prediction, target optional if truth provided
            tfunc_name = mixdb.target_file(1).truth_settings[0].function  # first target, assumes all have same
            if tfunc_name == 'mapped_snr_f':
                # leave as unmapped snr
                predplot = predict
                tfunc_name = mixdb.target_file(1).truth_settings[0].function
            elif tfunc_name == 'target_f' or 'target_mixture_f':
                predplot = 20 * np.log10(abs(predict_complex) + np.finfo(np.float32).eps)
            else:
                # use dB scale
                predplot = 10 * np.log10(predict + np.finfo(np.float32).eps)
                tfunc_name = tfunc_name + ' (db)'

            mixspec = 20 * np.log10(abs(mixture_f) + np.finfo(np.float32).eps)
            pdf.savefig(plot_mixpred(mixture=mixture,
                                     mixture_f=mixspec,
                                     target=target,
                                     feature=feat_sgram,
                                     predict=predplot,
                                     tp_title=tfunc_name))

            # ----- page 2, plot unmapped predict, opt truth reconstructed and line plots of mean-over-f
            # pdf.savefig(plot_pdb_predtruth(predict=pred_snr_f, tp_title='predict snr_f (db)'))

            # page 3 speech extraction
            tg_spec = 20 * np.log10(abs(target_f) + np.finfo(np.float32).eps)
            tg_est_spec = 20 * np.log10(abs(predict_complex) + np.finfo(np.float32).eps)
            # n_spec = np.reshape(n_spec,(n_spec.shape[0] * n_spec.shape[1], n_spec.shape[2]))
            pdf.savefig(plot_epredtruth(predict=tg_est_spec,
                                        predict_wav=target_est_wav,
                                        truth_f=tg_spec,
                                        truth_wav=targetfi,
                                        metric=np.vstack((lerr_tg_frame, phd_frame)).T,
                                        tp_title='speech estimate'))

            # page 4 noise extraction
            n_spec = 20 * np.log10(abs(noise_f) + np.finfo(np.float32).eps)
            n_est_spec = 20 * np.log10(abs(noise_est_complex) + np.finfo(np.float32).eps)
            pdf.savefig(plot_epredtruth(predict=n_est_spec,
                                        predict_wav=noise_est_wav,
                                        truth_f=n_spec,
                                        truth_wav=noisefi,
                                        metric=lerr_n_frame,
                                        tp_title='noise estimate'))

            # Plot error waveforms
            # tg_err_wav = targetfi - target_est_wav
            # tg_err_spec = 20*np.log10(np.abs(target_f - predict_complex))

        plt.close('all')

    return all_metrics_table_1, all_metrics_table_2


def main():
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    mixids = args['--mixid']
    predict_location = args['PLOC']
    wer_method = args['--wer-method'].lower()
    truth_est_mode = args['--truth-est-mode']
    enable_plot = args['--plot']
    enable_wav = args['--wav']
    enable_summary = args['--summary']
    truth_location = args['TLOC']
    whisper_model = args['--whisper-model'].lower()

    import glob
    from os.path import basename
    from os.path import isdir
    from os.path import join
    from os.path import split

    from tqdm import tqdm

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import logger
    from sonusai import update_console_handler
    from sonusai.mixture import DEFAULT_NOISE
    from sonusai.mixture import MixtureDatabase
    from sonusai.mixture import read_audio
    from sonusai.utils import calc_asr
    from sonusai.utils import pp_tqdm_imap

    # Check prediction subdirectory
    if not isdir(predict_location):
        print(f'The specified predict location {predict_location} is not a valid subdirectory path, exiting ...')

    # allpfiles = listdir(predict_location)
    allpfiles = glob.glob(predict_location + "/*.h5")
    predict_logfile = glob.glob(predict_location + "/*predict.log")
    predwav_mode = False
    if len(allpfiles) <= 0 and not truth_est_mode:
        allpfiles = glob.glob(predict_location + "/*.wav")  # check for wav files
        if len(allpfiles) <= 0:
            print(f'Subdirectory {predict_location} has no .h5 or .wav files, exiting ...')
        else:
            logger.info(f'Found {len(allpfiles)} prediction .wav files.')
            predwav_mode = True
    else:
        logger.info(f'Found {len(allpfiles)} prediction .h5 files.')

    if len(predict_logfile) == 0:
        logger.info(f'Warning, predict location {predict_location} has no prediction log files.')
    else:
        logger.info(f'Found predict log {basename(predict_logfile[0])} in predict location.')

    # Setup logging file
    create_file_handler(join(predict_location, 'calc_metric_spenh.log'))
    update_console_handler(verbose)
    initial_log_messages('calc_metric_spenh')

    mixdb = MixtureDatabase(truth_location)
    mixids = mixdb.mixids_to_list(mixids)
    logger.info(
        f'Found mixdb of {mixdb.num_mixtures} total mixtures, with {mixdb.num_classes} classes in {truth_location}')
    logger.info(f'Only running specified subset of {len(mixids)} mixtures')

    enable_asr_warmup = False
    if wer_method == 'none':
        fnb = 'metric_spenh_'
    elif wer_method == 'google':
        fnb = 'metric_spenh_ggl_'
        logger.info(f'WER enabled with method {wer_method}')
        enable_asr_warmup = True
    elif wer_method == 'deepgram':
        fnb = 'metric_spenh_dgram_'
        logger.info(f'WER enabled with method {wer_method}')
        enable_asr_warmup = True
    elif wer_method == 'aixplain_whisper':
        fnb = 'metric_spenh_whspx_' + whisper_model + '_'
        logger.info(f'WER enabled with method {wer_method} and whisper model {whisper_model}')
        enable_asr_warmup = True
    elif wer_method == 'whisper':
        fnb = 'metric_spenh_whspl_' + whisper_model + '_'
        logger.info(f'WER enabled with method {wer_method} and whisper model {whisper_model}')
        enable_asr_warmup = True
    elif wer_method == 'aaware_whisper':
        fnb = 'metric_spenh_whspaaw_' + whisper_model + '_'
        logger.info(f'WER enabled with method {wer_method} and whisper model {whisper_model}')
        enable_asr_warmup = True
    elif wer_method == 'fastwhisper':
        fnb = 'metric_spenh_fwhsp_' + whisper_model + '_'
        logger.info(f'WER enabled with method {wer_method} and whisper model {whisper_model}')
        enable_asr_warmup = True
    else:
        logger.error(f'Unrecognized WER method: {wer_method}')
        return

    if enable_asr_warmup:
        DEFAULT_SPEECH = split(DEFAULT_NOISE)[0] + '/speech_ma01_01.wav'
        audio = read_audio(DEFAULT_SPEECH)
        logger.info(f'Warming up asr method, note for cloud service this could take up to a few min ...')
        asr_chk = calc_asr(audio, engine=wer_method, whisper_model_name=whisper_model)
        logger.info(f'Warmup completed, results {asr_chk}')

    MP_GLOBAL.mixdb = mixdb
    MP_GLOBAL.predict_location = predict_location
    MP_GLOBAL.predwav_mode = predwav_mode
    MP_GLOBAL.truth_est_mode = truth_est_mode
    MP_GLOBAL.enable_plot = enable_plot
    MP_GLOBAL.enable_wav = enable_wav
    MP_GLOBAL.wer_method = wer_method
    MP_GLOBAL.whisper_model = whisper_model

    # Individual mixtures use pandas print, set precision to 2 decimal places
    # pd.set_option('float_format', '{:.2f}'.format)
    progress = tqdm(total=len(mixids), desc='calc_metric_spenh')
    all_metrics_tables = pp_tqdm_imap(_process_mixture, mixids, progress=progress, num_cpus=None)
    progress.close()

    all_metrics_table_1 = pd.concat([item[0] for item in all_metrics_tables])
    all_metrics_table_2 = pd.concat([item[1] for item in all_metrics_tables])

    if not enable_summary:
        return

    # 9) Done with mixtures, write out summary metrics
    # Calculate SNR summary avg of each non-random snr
    all_mtab1_sorted = all_metrics_table_1.sort_values(by=['MXSNR', 'SPFILE'])
    all_mtab2_sorted = all_metrics_table_2.sort_values(by=['MXSNR'])
    mtab_snr_summary = None
    mtab_snr_summary_em = None
    for snri in range(0, len(mixdb.snrs)):
        tmp = all_mtab1_sorted.query('MXSNR==' + str(mixdb.snrs[snri])).mean(numeric_only=True).to_frame().T
        # avoid nan when subset of mixids specified
        if ~np.isnan(tmp.iloc[0].to_numpy()[0]).any():
            mtab_snr_summary = pd.concat([mtab_snr_summary, tmp])

        tmp = all_mtab2_sorted[all_mtab2_sorted['MXSNR'] == mixdb.snrs[snri]].mean(numeric_only=True).to_frame().T
        # avoid nan when subset of mixids specified (mxsnr will be nan if no data):
        if ~np.isnan(tmp.iloc[0].to_numpy()[0]).any():
            mtab_snr_summary_em = pd.concat([mtab_snr_summary_em, tmp])

    # Correct percentages in snr summary table
    mtab_snr_summary['PESQi%'] = 100 * (mtab_snr_summary['PESQ'] - mtab_snr_summary['MXPESQ']) / np.maximum(
        mtab_snr_summary['MXPESQ'], 0.01)
    for i in range(len(mtab_snr_summary)):
        if mtab_snr_summary['MXWER'].iloc[i] == 0.0:
            if mtab_snr_summary['WER'].iloc[i] == 0.0:
                mtab_snr_summary['WERi%'].iloc[i] = 0.0
            else:
                mtab_snr_summary['WERi%'].iloc[i] = -999.0
        else:
            mtab_snr_summary['WERi%'].iloc[i] = 100 * (mtab_snr_summary['MXWER'].iloc[i] -
                                                       mtab_snr_summary['WER'].iloc[i]) / \
                                                mtab_snr_summary['MXWER'].iloc[i]

    # Calculate avg metrics over all mixtures except -99
    all_mtab1_sorted_nom99 = all_mtab1_sorted[all_mtab1_sorted.MXSNR != -99]
    all_nom99_mean = all_mtab1_sorted_nom99.mean(numeric_only=True)

    # correct the percentage averages with a direct calculation (PESQ% and WER%):
    # ser.iloc[pos]
    all_nom99_mean['PESQi%'] = (100 * (all_nom99_mean['PESQ'] - all_nom99_mean['MXPESQ'])
                                / np.maximum(all_nom99_mean['MXPESQ'], 0.01))  # pesq%
    # all_nom99_mean[3] = 100 * (all_nom99_mean[2] - all_nom99_mean[1]) / np.maximum(all_nom99_mean[1], 0.01)  # pesq%
    if all_nom99_mean['MXWER'] == 0.0:
        if all_nom99_mean['WER'] == 0.0:
            all_nom99_mean['WERi%'] = 0.0
        else:
            all_nom99_mean['WERi%'] = -999.0
    else:  # wer%
        all_nom99_mean['WERi%'] = 100 * (all_nom99_mean['MXWER'] - all_nom99_mean['WER']) / all_nom99_mean['MXWER']

    num_mix = len(mixids)
    if num_mix > 1:
        # Print pandas data to files using precision to 2 decimals
        # pd.set_option('float_format', '{:.2f}'.format)
        csp = 0

        if not truth_est_mode:
            ofname = join(predict_location, fnb + 'summary.txt')
        else:
            ofname = join(predict_location, fnb + 'summary_truest.txt')

        with open(ofname, 'w') as f:
            print(f'WER enabled with method {wer_method}, whisper model, if used: {whisper_model}', file=f)
            print(f'Speech enhancement metrics avg over all {len(all_mtab1_sorted_nom99)} non -99 SNR mixtures:',
                  file=f)
            print(all_nom99_mean.to_frame().T.round(2).to_string(float_format=lambda x: "{:.2f}".format(x),
                                                                 index=False), file=f)
            print(f'\nSpeech enhancement metrics avg over each SNR:', file=f)
            print(mtab_snr_summary.round(2).to_string(float_format=lambda x: "{:.2f}".format(x), index=False), file=f)
            print('', file=f)
            print(f'Extraction statistics stats avg over each SNR:', file=f)
            # with pd.option_context('display.max_colwidth', 9):
            # with pd.set_option('float_format', '{:.1f}'.format):
            print(mtab_snr_summary_em.round(1).to_string(float_format=lambda x: "{:.1f}".format(x), index=False),
                  file=f)
            print('', file=f)
            # pd.set_option('float_format', '{:.2f}'.format)

            print(f'Speech enhancement metrics stats over all {num_mix} mixtures:', file=f)
            print(all_metrics_table_1.describe().round(2).to_string(float_format=lambda x: "{:.2f}".format(x)), file=f)
            print('', file=f)
            print(f'Extraction statistics stats over all {num_mix} mixtures:', file=f)
            print(all_metrics_table_2.describe().round(2).to_string(float_format=lambda x: "{:.1f}".format(x)), file=f)
            print('', file=f)

            print('Speech enhancement metrics all-mixtures list:', file=f)
            # print(all_metrics_table_1.head().style.format(precision=2), file=f)
            print(all_metrics_table_1.round(2).to_string(float_format=lambda x: "{:.2f}".format(x)), file=f)
            print('', file=f)
            print('Extraction statistics all-mixtures list:', file=f)
            print(all_metrics_table_2.round(2).to_string(float_format=lambda x: "{:.1f}".format(x)), file=f)

        # Write summary to .csv file
        if not truth_est_mode:
            csv_name = join(predict_location, fnb + 'summary.csv')
        else:
            csv_name = join(predict_location, fnb + 'summary_truest.csv')
        header_args = {
            'mode':     'a',
            'encoding': 'utf-8',
            'index':    False,
            'header':   False,
        }
        table_args = {
            'mode':     'a',
            'encoding': 'utf-8',
        }
        label = f'Speech enhancement metrics avg over all {len(all_mtab1_sorted_nom99)} non -99 SNR mixtures:'
        pd.DataFrame([label]).to_csv(csv_name, header=False, index=False)  # open as write
        all_nom99_mean.to_frame().T.round(2).to_csv(csv_name, index=False, **table_args)
        pd.DataFrame(['']).to_csv(csv_name, **header_args)
        pd.DataFrame([f'Speech enhancement metrics avg over each SNR:']).to_csv(csv_name, **header_args)
        mtab_snr_summary.round(2).to_csv(csv_name, index=False, **table_args)
        pd.DataFrame(['']).to_csv(csv_name, **header_args)
        pd.DataFrame([f'Extraction statistics stats avg over each SNR:']).to_csv(csv_name, **header_args)
        mtab_snr_summary_em.round(2).to_csv(csv_name, index=False, **table_args)
        pd.DataFrame(['']).to_csv(csv_name, **header_args)
        pd.DataFrame(['']).to_csv(csv_name, **header_args)
        label = f'Speech enhancement metrics stats over {num_mix} mixtures:'
        pd.DataFrame([label]).to_csv(csv_name, **header_args)
        all_metrics_table_1.describe().round(2).to_csv(csv_name, **table_args)
        pd.DataFrame(['']).to_csv(csv_name, **header_args)
        label = f'Extraction statistics stats over {num_mix} mixtures:'
        pd.DataFrame([label]).to_csv(csv_name, **header_args)
        all_metrics_table_2.describe().round(2).to_csv(csv_name, **table_args)
        label = f'WER enabled with method {wer_method}, whisper model, if used: {whisper_model}'
        pd.DataFrame([label]).to_csv(csv_name, **header_args)

        if not truth_est_mode:
            csv_name = join(predict_location, fnb + 'list.csv')
        else:
            csv_name = join(predict_location, fnb + 'list_truest.csv')
        pd.DataFrame(['Speech enhancement metrics list:']).to_csv(csv_name, header=False, index=False)  # open as write
        all_metrics_table_1.round(2).to_csv(csv_name, **table_args)

        if not truth_est_mode:
            csv_name = join(predict_location, fnb + 'estats_list.csv')
        else:
            csv_name = join(predict_location, fnb + 'estats_list_truest.csv')
        pd.DataFrame(['Extraction statistics list:']).to_csv(csv_name, header=False, index=False)  # open as write
        all_metrics_table_2.round(2).to_csv(csv_name, **table_args)


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        exit()



from sonusai.mixture.datatypes import AudioT
from sonusai.mixture.datatypes import Truth
from sonusai.mixture.datatypes import TruthFunctionConfig
from sonusai.mixture.mixdb import MixtureDatabase


def truth_function(target_audio: AudioT,
                   noise_audio: AudioT,
                   mixture_audio: AudioT,
                   config: TruthFunctionConfig) -> Truth:
    from sonusai import SonusAIError
    from sonusai.mixture import truth_functions
    from .truth_functions.data import Data

    data = Data(target_audio, noise_audio, mixture_audio, config)
    if data.config.target_gain == 0:
        return data.truth

    try:
        return getattr(truth_functions, data.config.function)(data)
    except AttributeError:
        raise SonusAIError(f'Unsupported truth function: {data.config.function}')


def get_truth_indices_for_mixid(mixdb: MixtureDatabase, mixid: int) -> list[int]:
    """Get a list of truth indices for a given mixid."""
    from .targets import get_truth_indices_for_target

    indices: list[int] = []
    for target_id in [target.file_id for target in mixdb.mixture(mixid).targets]:
        indices.append(*get_truth_indices_for_target(mixdb.target_file(target_id)))

    return sorted(list(set(indices)))


def truth_reduction(x: Truth, func: str) -> Truth:
    import numpy as np

    from sonusai import SonusAIError

    if func == 'max':
        return np.max(x, axis=0)

    if func == 'mean':
        return np.mean(x, axis=0)

    if func == 'index0':
        return np.squeeze(x[0, :])

    raise SonusAIError(f'Invalid truth reduction function: {func}')
